# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..pyzef.main import *
from ..pyzef import main, zefops, internals
from ._core import ET, AtomicEntityType

main.ZefRef.__hash__ = lambda self: hash(((index(self)), index(self | zefops.tx)))
main.EZefRef.__hash__ = lambda self: index(self)

main.QuantityFloat.__hash__ = lambda self: hash((self.value, self.unit))
main.QuantityInt.__hash__ = lambda self: hash((self.value, self.unit))
main.Time.__hash__ = lambda self: hash(self.seconds_since_1970)

internals.Delegate.__hash__ = lambda self: hash((self.order, self.item))
internals.DelegateEntity.__hash__ = lambda self: hash(self.et)
internals.DelegateAtomicEntity.__hash__ = lambda self: hash(self.aet)
internals.DelegateRelationGroup.__hash__ = lambda self: hash(self.rt)
internals.DelegateRelationTriple.__hash__ = lambda self: hash((self.rt, self.source, self.target))
internals.DelegateTX.__hash__ = lambda self: hash(internals.DelegateTX)
internals.DelegateRoot.__hash__ = lambda self: hash(internals.DelegateRoot)



# FIXME: why do we need the standard list of timezones in the docstring for Time? Should just refer people to the TZ database. I have removed it for now.

# ------------------------- patch Time constructor to allow create of zef Time objects both from time stamps and readable strings ------------------------
from datetime import datetime, timezone, timedelta
import dateparser
import pytz   # for time zones

def create_monkey_patched_Time_ctor(old_init_fct):
    def monkey_patched_Time_ctor(self, x, timezone: str = 'Asia/Singapore'):
        """ x could be a time stamp (float/int) or a string to be parsed
        Note: the python library date parser parses dates of the following form UNINTUITIVELY, i.e. it assumes the MDY convention:  
            '9.2.2015 13:59' is read as Sep. 2. 2015
            Hence: ALWAYS stick to the convention of transforming strings to the form Time('2019-07-15 13:59', timezone='Europe/London')

            Other formats work, e.g.
            Time('9. Feb 2015 13:59')
            Time('Martes 21 de Octubre de 2014 13:23:12')    # various languages
            Time('1 เดือนตุลาคม 2005, 1:00 AM')    # Thai
            Time('1 hour ago')         # relative dates
            Time('now')

        The time zone can be specified as a string: 
            t0 = Time('2019-07-15 13:59', timezone='Europe/Berlin')
        If no time zone is specified, Singapore is assumed as a default for now - this will be cleaned up in the future.
        

        Valid timezones: see TZ database.
        """
        if(isinstance(x, float)):
            self = old_init_fct(self, x)
        elif(isinstance(x, int)):
            self = old_init_fct(self, float(x))
        elif(isinstance(x, zefops.Now)):
            self = old_init_fct(self, x)
        elif(isinstance(x, str)):
            try:
                my_timestamp = datetime.timestamp(dateparser.parse(x, settings={'TIMEZONE': timezone, 'RETURN_AS_TIMEZONE_AWARE': True}))
                self = old_init_fct(self, my_timestamp)
            except:
                raise ValueError(f"Could not parse the following string as a valid datetime: \"{x}\"")
        else:
            raise ValueError('Invalid argument type passed to Time constructor')
    return monkey_patched_Time_ctor
main.Time.__init__ = create_monkey_patched_Time_ctor(Time.__init__)


def _to_date_time(self: Time, timezone:str='Asia/Singapore', str_format: str='%Y-%m-%d %H:%M:%S'):
    return datetime.fromtimestamp(self.seconds_since_1970)\
            .astimezone(pytz.timezone(timezone))\
            .strftime(str_format)

def _to_time(self: Time, timezone:str='Asia/Singapore'):
    return _to_date_time(self, timezone, str_format='%H:%M:%S')

def _to_date(self: Time, timezone:str='Asia/Singapore'):
    return _to_date_time(self, timezone, str_format='%Y-%m-%d')



main.Time.date_time = _to_date_time
main.Time.time = _to_time
main.Time.date = _to_date


def _repr_for_zef_time(self):
    return datetime.fromtimestamp(self.seconds_since_1970)\
                .astimezone(timezone(timedelta(hours=8)))\
                .strftime(f"<Time %Y-%m-%d %H:%M:%S (+0800)>")
def _str_for_zef_time(self):
    return datetime.fromtimestamp(self.seconds_since_1970)\
                .astimezone(timezone(timedelta(hours=8)))\
                .strftime(f"%Y-%m-%d %H:%M:%S (+0800)")
main.Time.__repr__ = _repr_for_zef_time
main.Time.__str__ = _str_for_zef_time


def help(arg):
    import builtins
    from . import ZefRef
    from ..zefops.base import L, value
    if isinstance(arg, ZefRef) and arg | BT == BT.ENTITY_NODE and arg | ET == ET.ZEF_Function:
        if len(arg >> L[RT.DocString]) == 1:
            print((arg >> RT.DocString) | value.String)
        else:
            print("This ZEF_Function doesn't have any DocString")
    else:
        print(builtins.help(arg))



# Temp fix fow now: these operators should be overloaded in C++ and return a new C++ type.
def _rae_get_item(self, n):
    """
    we want to write  RT.UsedBy['some internal id']. 
    Same for ETs, AETs. Monkeypatch that from Python for now.
    """
    from ._ops import instantiated
    # return {'RAE': self, 'capture': n}
    return instantiated[self][n]




main.QuantityFloat.__getitem__ = _rae_get_item
main.QuantityInt.__getitem__ = _rae_get_item



# ---------------------------- special treatment for EntityType ------------------------------------

def absorbed_get_item(self, x):
    # don't mutate, create a new instance
    from copy import copy
    new_obj = copy(self)
    if '_absorbed' not in self.__dict__:
        new_obj._absorbed = (x, )
    else:
        new_obj._absorbed = (*self._absorbed, x)
    return new_obj
    

def repr_with_absorbed(self, orig_repr):
    original = orig_repr(self)
    if '_absorbed' not in self.__dict__:
        return original
    else:
        return original + ''.join(('[' + repr(el) + ']' for el in self._absorbed))

def wrap_repr(typ):
    orig = typ.__repr__
    typ.__repr__ = lambda self,orig=orig: repr_with_absorbed(self, orig)
    


# The C++ implementation does not consider the absorbed values.
# Monkeypatch this function here.
def keyword_eq_monkeypatched(self, other):
    if not isinstance(other, main.Keyword):
        return False
    if self.value != other.value:
        return False
    return self.__dict__.get('_absorbed', None) == other.__dict__.get('_absorbed', None) 



main.EntityType.__getitem__ = absorbed_get_item
wrap_repr(main.EntityType)

main.RelationType.__getitem__ = absorbed_get_item
wrap_repr(main.RelationType)

main.Keyword.__getitem__ = absorbed_get_item
wrap_repr(main.Keyword)
main.Keyword.__eq__ = keyword_eq_monkeypatched

internals.AtomicEntityType.__getitem__ = absorbed_get_item
wrap_repr(internals.AtomicEntityType)

internals.Delegate.__getitem__ = absorbed_get_item
wrap_repr(internals.Delegate)






# we want to allow for e.g. "AET.Float <= 42.1"
# return a ZefOp to use as a container here.
# We used a dict previously, but then we could not 
# catch the case "AET.Float['abc'] <= 42.1"
# For consistency, we use Lazy ZefOps as intermediate containers
def leq_monkey_patching_ae(self, other):
    from ._ops import assign_value, instantiated, LazyValue
    return LazyValue(self) | assign_value[other]
    
AtomicEntityType.__le__ = leq_monkey_patching_ae


# Pretty printing for ZefRefs
def pprint_ZefRefs(self, p, cycle):
    p.text(str(self))
    p.text(" [\n")

    N_max = 10
    if len(self) > N_max:
        # Split into first part and last part
        N = N_max // 2
        first = [self[i] for i in range(N)]
        last = [self[i] for i in range(len(self)-N, len(self)-1)]
        p.text('\n'.join("\t"+str(x) for x in first))
        p.text('\n\t...\n')
        p.text('\n'.join("\t"+str(x) for x in last))
    else:
        p.text('\n'.join("\t"+str(x) for x in self))
    p.text(']')

main.ZefRefs._repr_pretty_ = pprint_ZefRefs
main.EZefRefs._repr_pretty_ = pprint_ZefRefs

def convert_to_assign_value(self, value):
    from ._ops import assign_value
    return self | assign_value[value]
ZefRef.__le__ = convert_to_assign_value
EZefRef.__le__ = convert_to_assign_value

def add_internal_id(self, internal_id):
    from ._ops import merged
    return merged[self][internal_id]
ZefRef.__getitem__ = add_internal_id
EZefRef.__getitem__ = add_internal_id

original_Graph__contains__ = main.Graph.__contains__
def Graph__contains__(self, x):
    from .abstract_raes import Entity, AtomicEntity, Relation
    from ._ops import origin_uid
    if type(x) in [Entity, AtomicEntity, Relation]:
        return origin_uid(x) in self

    if type(x) in [ZefRef, EZefRef]:
        if Graph(x) == self:
            return True
        return origin_uid(x) in self

    return original_Graph__contains__(self, x)
main.Graph.__contains__ = Graph__contains__
    
original_Graph__getitem__ = main.Graph.__getitem__
def Graph__getitem__(self, x):
    from .abstract_raes import Entity, AtomicEntity, Relation
    from ._ops import uid, target
    from .internals import BT
    if type(x) in [Entity, AtomicEntity, Relation]:
        return self[uid(x)]

    res = original_Graph__getitem__(self, x)
    # We magically transform any FOREIGN_ENTITY_NODE accesses to the real RAEs.
    # Accessing the low-level BTs can only be done through traversals
    if BT(res) in [BT.FOREIGN_ENTITY_NODE, BT.FOREIGN_ATOMIC_ENTITY_NODE, BT.FOREIGN_RELATION_EDGE]:
        # return target(res << BT.ORIGIN_RAE_EDGE)
        # TODO: We need a consistent way of accessing this.
        #
        # Is this even possible for a EZefRef? User can terminate and recreate
        # RAEs that have the same identity, so there's nothing sensible that
        # could be returned here?
        #
        # Hence, going to disable this. It will be that you can only get the RAE
        # corresponding to a foreign RAE in the context of a graph slice. Doing
        # it here will just return the low level blob.
        pass
    return res

main.Graph.__getitem__ = Graph__getitem__



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    "config"
]

import os
from pathlib import Path
import yaml
from ..core.op_implementations.dispatch_dictionary import _op_to_functions
from ..core import *
from ..core._ops import *

from dataclasses import dataclass
import typing
@dataclass
class ConfigItem:
    path: str
    # These two types will have to be handled later on
    typ: typing.Any
    default: typing.Any
    env_var: str
    doc: str
    

config_spec = [
    # Options are "false", "auto", "always"
    ConfigItem("login.autoConnect", None, "auto", "ZEFDB_LOGIN_AUTOCONNECT", "Whether zefdb should automatically connect to ZefHub on start of the butler."),
    # In the future, this should be true not "true"
    ConfigItem("butler.autoStart", bool, "true", "ZEFDB_BUTLER_AUTOSTART", "Whether the butler will automatically be started on import of the zefdb module."),
]
config_spec = {x.path: x for x in config_spec}
# Anything that has been manually set in this session will override any environment variable etc...
session_overrides = {}


config_file_path = Path.home() / ".zef" / "config.yaml"


def config_file_exists():
    return os.path.exists(config_file_path)

def ensure_config_file():
    if os.path.exists(config_file_path):
        return
    config_file_path.parent.mkdir(parents=True, exist_ok=True)
    # if not os.access(str(config_file_path), os.W_OK):
    #     raise Exception(f"Don't have permissions to write to config file path: {config_file_path}") 
    with open(config_file_path, 'w'):
        pass

def nested_set(dic, keys, value):
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})
    dic[keys[-1]] = value


def get_config(key=""):
    # TODO: Danny did this quickly, but it would only work for specific keys.
    # Need to handle wildcards/group selections.
    if key in session_overrides:
        return session_overrides[key]
    if key in config_spec:
        spec = config_spec[key]
        import os
        if spec.env_var in os.environ:
            # TODO: Handle conversion from string here
            return os.environ[spec.env_var]

    # Check if config file exists
    if config_file_exists():
        with open(config_file_path, 'r') as stream:
            try:
                config = yaml.safe_load(stream)
                if config is None:
                    # TODO: Handle warnings about improper config here
                    return None

                # Check if key is empty or wildcard
                if key == "" or key == "*":
                    return config

                # Check if key is in config and return value
                for k in key.split("."):
                    if k not in config:
                        config = None
                        break
                    config = config[k]
                if config is not None:
                    return config
            except yaml.YAMLError as exc:
                print(exc)

    # This is the default option - we also assume that by getting here we must
    # have a valid config key, so error if nothing matches.
    if key not in config_spec:
        raise Exception(f"Unknown config option {key}")

    return config_spec[key].default
        
    


def set_config(key, val):
    ensure_config_file()
    try:
        with open(config_file_path, 'r') as stream:
            config = yaml.safe_load(stream)
            if key == "":
                return config
            if config is None:
                config = {}

            nested_set(config, key.split("."), val)
            yamlConfig = yaml.safe_dump(config, default_flow_style=False)
        with open(config_file_path, 'w') as filehandle:
            filehandle.write(yamlConfig)
    except yaml.YAMLError as exc:
        print(exc)

    if key in config_spec:
        session_overrides[key] = val


def list_config():
    if config_file_exists():
        with open(config_file_path, 'r') as stream:
            try:
                config = yaml.safe_load(stream)
                # TODO: Print ascii table?
            except yaml.YAMLError as exc:
                print(exc)


def config_implementation(payload, action):
    if action == KW.set:
        assert isinstance(payload, tuple)
        assert len(payload) == 2
        assert isinstance(payload[0], str)
        set_config(payload[0], payload[1])
    elif action == KW.get:
        assert isinstance(payload, str)
        return get_config(payload)
    else:
        raise NotImplementedError("Action {} not implemented".format(action))


def config_typeinfo(v_tp):
    return VT.Any


_op_to_functions[RT.Config] = (config_implementation, config_typeinfo)

config = ZefOp(((RT.Config, ()), ))



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Assuming this file is not imported during zefdb init.
from ... import *
from ...ops import *
from functools import partial as P
from ...core.logger import log

from ariadne import ObjectType, QueryType, MutationType, EnumType, ScalarType

##############################
# * Utils
#----------------------------

@func
def is_core_scalar(z):
    return (
        is_a(z, AET.String) or
        is_a(z, AET.Int) or
        is_a(z, AET.Float) or
        is_a(z, AET.Bool)
    )
    
assert_type = Assert[Or[is_a[ET.GQL_Type]][is_a[ET.GQL_Enum]][is_a[AET]]][lambda z: f"{z} is not a GQL type"]
assert_field = Assert[is_a[RT.GQL_Field]][lambda z: f"{z} is not a GQL field"]

@func
def single_or(itr, default):
    itr = iter(itr)
    try:
        ret = next(itr)
        try:
            next(itr)
            raise Exception("single_or detected more than one item in iterator")
        except StopIteration:
            return ret
    except StopIteration:
        return default
        
op_is_scalar = assert_type | Or[is_a[AET]][is_a[ET.GQL_Enum]]
op_is_orderable = assert_type | Or[is_a[AET.Float]][is_a[AET.Int]][is_a[AET.Time]]
op_is_summable = assert_type | Or[is_a[AET.Float]][is_a[AET.Int]]
op_is_stringlike = assert_type | is_a[AET.String]
op_is_list = assert_field >> O[RT.List] | value_or[False] | collect
op_is_required = assert_field >> O[RT.Required] | value_or[False] | collect
op_is_unique = assert_field >> O[RT.Unique] | value_or[False] | collect
op_is_searchable = assert_field >> O[RT.Search] | value_or[False] | collect
op_is_aggregable = assert_field | And[Not[op_is_list]][target | Or[op_is_orderable][op_is_summable]]
op_is_incoming = assert_field >> O[RT.Incoming] | value_or[False] | collect

op_is_upfetch = assert_field >> O[RT.Upfetch] | value_or[False] | collect
op_upfetch_field = (assert_type > L[RT.GQL_Field]
                    | filter[op_is_upfetch]
                    | single_or[None]
                    | collect)
op_has_upfetch = op_upfetch_field | Not[equals[None]] | collect

########################################
# * Generating it all
#--------------------------------------

def generate_resolvers_fcts(schema_root):
    Query = QueryType()
    Mutation = MutationType()

    all_objects = [Query, Mutation]
    query_fields = []
    mutation_fields = []
    type_schemas = []
    enum_schemas = []

    extra_filters = {}


    TimeScalar = ScalarType("DateTime")
    all_objects += [TimeScalar]
    import datetime
    TimeScalar.set_serializer(lambda t: datetime.datetime.fromtimestamp(t.seconds_since_1970).isoformat())
    TimeScalar.set_value_parser(Time)

    for z_type in schema_root >> L[RT.GQL_Type]:
        name = value(z_type >> RT.Name)
        Type = ObjectType(name)
        all_objects += [Type]

        field_schemas = []
        ref_field_schemas = []
        add_field_schemas = []
        upfetch_field_schemas = []
        aggregate_outs = []

        for z_field in z_type > L[RT.GQL_Field]:
            field_name = z_field >> RT.Name | value | collect
            Type.set_field(field_name,
                            P(resolve_field, z_field=z_field))

            is_scalar = z_field | target | op_is_scalar | collect
            is_required = z_field | op_is_required | collect
            is_list = z_field | op_is_list | collect
            is_orderable = z_field | target | op_is_orderable | collect
            is_summable = z_field | target | op_is_orderable | collect
            is_aggregable = z_field | op_is_aggregable | collect

            if op_is_unique(z_field):
                assert is_scalar, "Unique fields must be scalars"
                assert not is_list, "Unique fields can't be lists"

            if op_is_upfetch(z_field):
                assert op_is_unique(z_field), "Upfetch field must be unique"
                assert is_required, "Upfetch field must be required"

            base_field_type = z_field | target >> RT.Name | value | collect
            if is_scalar:
                base_ref_field_type = base_field_type
            else:
                base_ref_field_type = base_field_type + "Ref" 

            if is_required:
                maybe_required = "!"
            else:
                maybe_required = ""
            if is_list:
                maybe_params = "(" + schema_generate_list_params(target(z_field), extra_filters) + ")"
                ref_field_type = "[" + base_ref_field_type + "]"
                # add_field_type = "[" + base_ref_field_type + maybe_required + "]"
                add_field_type = ref_field_type
                upfetch_field_type = "[" + base_ref_field_type + maybe_required + "]"
                field_type = "[" + base_field_type + maybe_required + "]"
            else:
                maybe_params = ""
                ref_field_type = base_ref_field_type
                # add_field_type = base_ref_field_type + maybe_required
                add_field_type = ref_field_type
                upfetch_field_type = base_ref_field_type + maybe_required
                field_type = base_field_type + maybe_required
            field_schemas += [f"{field_name}{maybe_params}: {field_type}"]
            add_field_schemas += [f"{field_name}: {add_field_type}"]
            ref_field_schemas += [f"{field_name}: {ref_field_type}"]
            upfetch_field_schemas += [f"{field_name}: {upfetch_field_type}"]

            if is_aggregable:
                if is_orderable:
                    aggregate_outs += [
                        f"{field_name}Min: {ref_field_type}",
                        f"{field_name}Max: {ref_field_type}",
                    ]
                if is_summable:
                    aggregate_outs += [f"{field_name}Sum: {ref_field_type}"]
                    aggregate_outs += [f"{field_name}Avg: {ref_field_type}"]

        field_schemas += ["id: ID!"]
        add_field_schemas += ["id: ID"]
        ref_field_schemas += ["id: ID"]
        # Note: upfetch does not get an id field
        Type.set_field("id", resolve_id)

        # Add the 3 top-level queries
        Query.set_field(f"get{name}",
                        P(resolve_get, type_node=z_type))
        Query.set_field(f"query{name}",
                        P(resolve_query, type_node=z_type))
        Query.set_field(f"aggregate{name}",
                        P(resolve_aggregate, type_node=z_type))
        # Add the 3 top-level mutations
        Mutation.set_field(f"add{name}",
                            P(resolve_add, type_node=z_type))
        Mutation.set_field(f"update{name}",
                            P(resolve_update, type_node=z_type))
        Mutation.set_field(f"delete{name}",
                            P(resolve_delete, type_node=z_type))
        has_upfetch = op_has_upfetch(z_type)
        if has_upfetch:
            Mutation.set_field(f"upfetch{name}",
                               P(resolve_upfetch, type_node=z_type))

        MutateResponse = ObjectType(f"Mutate{name}Response")
        all_objects += [MutateResponse]

        # MutateResponse.set_field(f"count", lambda x,*args,**kwds: x["count"])
        MutateResponse.set_field(to_camel_case(name), P(resolve_filter_response, type_node=z_type))

        query_params = schema_generate_list_params(z_type, extra_filters)
        field_schemas = '\n\t'.join(field_schemas)
        ref_field_schemas = '\n\t'.join(ref_field_schemas)
        add_field_schemas = '\n\t'.join(add_field_schemas)
        upfetch_field_schemas = '\n\t'.join(upfetch_field_schemas)
        aggregate_outs = '\n\t'.join(aggregate_outs)

        type_schemas += [f"""type {name} {{
        {field_schemas}
}}

input {name}Ref {{
        {ref_field_schemas}
}}

input Add{name}Input {{
        {add_field_schemas}
}}
        
input Update{name}Input {{
        filter: {name}Filter!
        set: {name}Ref
        remove: {name}Ref
}}

type Mutate{name}Response {{
        {to_camel_case(name)}({query_params}): [{name}]
        count: Int
}}

type Aggregate{name}Response {{
        count: Int
        {aggregate_outs}
}}

"""]
        if has_upfetch:
            type_schemas += [f"""
input Upfetch{name}Input {{
        {upfetch_field_schemas}
}}
"""]

        # id_fields = z_type > L[RT.GQL_Field] | filter[Z >> O[RT.IsID] | value_or[False]]

        # get_params = id_fields | map[lambda z: f"{value(z >> RT.Name)}: {value(z | target >> RT.Name)}"] | collect
        # get_params = ", ".join(get_params)
        get_params = "id: ID!"
        query_fields += [
            f"get{name}({get_params}): {name}",
            f"query{name}({query_params}): [{name}]",
            f"aggregate{name}({query_params}): Aggregate{name}Response",
        ]
        mutation_fields += [
            f"add{name}(input: [Add{name}Input!]!, upsert: Boolean): Mutate{name}Response",
            f"update{name}(input: Update{name}Input!): Mutate{name}Response",
            f"delete{name}(filter: {name}Filter!): Mutate{name}Response",
        ]
        if has_upfetch:
            mutation_fields += [
                f"upfetch{name}(input: [Upfetch{name}Input!]!): Mutate{name}Response",
            ]

    for z_enum in schema_root >> L[RT.GQL_Enum]:
        name = value(z_enum >> RT.Name)

        opts = {}
        for z_opt in z_enum >> L[RT.GQL_Field]:
            assert is_a(z_opt, AET.Enum(name))
            opt_en = value(z_opt)
            opts[opt_en.enum_value] = opt_en

        Enum = EnumType(name, opts)

        opt_list = '\n\t'.join(opts.keys())

        enum_schemas += [f"""enum {name} {{
\t{opt_list}
}}"""]

        all_objects += [Enum]

    # Always generate the Int scalar type 
    int_type = schema_root >> L[RT.GQL_CoreScalarType] | filter[is_a[AET.Int]] | single | collect
    if int_type not in [rae_type(x) for x in extra_filters.keys()]:
        extra_filters[int_type] = schema_generate_scalar_filter(int_type)

    query_fields = '\n\t'.join(query_fields)
    mutation_fields = '\n\t'.join(mutation_fields)
    type_schemas = '\n\n'.join(type_schemas)
    enum_schemas = '\n\n'.join(enum_schemas)
    extra_filters = '\n\n'.join(x["schema"] for x in extra_filters.values())
    schema = f"""
{type_schemas}

{enum_schemas}

type Query {{
\t{query_fields}
}}

type Mutation {{
    {mutation_fields}
}}
    
scalar DateTime
    
{extra_filters}
    
"""

    return schema, all_objects

################################################
# * Schema specific parts
#----------------------------------------------

def schema_generate_list_params(z_type, extra_filters):
    name = value(z_type >> RT.Name)
    if z_type not in extra_filters:
        extra_filters[z_type] = None
        extra_filters[z_type] = schema_generate_type_filter(z_type, extra_filters)

    query_params = [
        f"filter: {name}Filter",
        # We probably want to change these to be proper cursors.
        "first: Int",
        "offset: Int",
    ]
    if extra_filters[z_type]["orderable"]:
        query_params += [f"order: {name}Order"]

    query_params = ", ".join(query_params)
    return query_params

def schema_generate_type_filter(z_type, extra_filters):
    name = value(z_type >> RT.Name)
    fil_name = f"{name}Filter"
    list_fil_name = f"{name}FilterList"
    order_name = f"{name}Order"
    orderable_name = f"{name}Orderable"

    fields = [
        # TODO: I Think this is if the field is present?
        # "has: [{fil_name}]",
        f"and: [{fil_name}]",
        f"or: [{fil_name}]",
        f"not: {fil_name}",
    ]

    orderable_fields = []

    # Every type has an ID field, which is slightly custom - functions as an
    # automatically generated "in"
    fields += [f"id: [ID!]"]

    for field in z_type > L[RT.GQL_Field] | filter[op_is_searchable]:
        field_name = value(field >> RT.Name)
        field_type = target(field)
        field_type_name = value(field_type >> RT.Name)
        if is_a(field_type, AET.Bool):
            filter_name = Boolean
        else:
            filter_name = f"{field_type_name}Filter"
            if field_type not in extra_filters:
                extra_filters[field_type] = None
                if op_is_scalar(field_type):
                    extra_filters[field_type] = schema_generate_scalar_filter(field_type)
                else:
                    extra_filters[field_type] = schema_generate_type_filter(field_type, extra_filters)

        if op_is_list(field):
            fields += [f"{field_name}: {filter_name}List"]
        else:
            fields += [f"{field_name}: {filter_name}"]

    for field in z_type > L[RT.GQL_Field] | filter[target | op_is_orderable]:
        orderable_fields += [value(field >> RT.Name)]

    fields = "\n\t".join(fields)
    orderable_fields = "\n\t".join(orderable_fields)

    out = f"""input {fil_name} {{
\t{fields}
}}
"""
    if len(orderable_fields) > 0:
        out += f"""
input {order_name} {{
        asc: {orderable_name}
        desc: {orderable_name}
        then: {order_name}
}}
        
input {list_fil_name} {{
\tany: {fil_name}
\tall: {fil_name}
\tsize: IntFilter
}}
    
enum {orderable_name} {{
\t{orderable_fields}
}}
"""
    return {"schema": out,
            "orderable": len(orderable_fields) > 0}
        

def schema_generate_scalar_filter(z_node):
    type_name = z_node >> RT.Name | value | collect
    fil_name = f"{type_name}Filter"
    list_fil_name = f"{type_name}FilterList"

    if is_a(z_node, AET.Bool):
        # Shoudln't need to do anything here, as it is true/false.
        return

    schema = f"""input {fil_name} {{
\teq: {type_name}
\tin: [{type_name}]
"""

    if op_is_orderable(z_node):
        schema += f"""\tle: {type_name}
\tlt: {type_name}
\tge: {type_name}
\tgt: {type_name}
\tbetween: {type_name}Range
"""

    if op_is_stringlike(z_node):
        schema += f"""\tcontains: String
"""

    schema += f"""}}

input {list_fil_name} {{
\tany: {fil_name}
\tall: {fil_name}
\tsize: IntFilter
}}
"""
    
    if op_is_orderable(z_node):
        schema += f"""\ninput {type_name}Range {{
\tmin: {type_name}!
\tmax: {type_name}!
}}"""

    return {"schema": schema}

####################################
# * Query resolvers
#----------------------------------

class ExternalError(Exception):
    pass

def resolve_get(_, info, *, type_node, **params):
    # Look for something that fits exactly what has been given in the params, assuming
    # that ariadne has done its work and validated the query.
    return find_existing_entity_by_id(info, type_node, params["id"])

def resolve_query(_, info, *, type_node, **params):
    g = info.context["g"]
    ents = g | now | all[ET(type_node >> RT.GQL_Delegate | collect)]
    ents = ents | filter[pass_query_auth[type_node][info]]

    ents = handle_list_params(ents, type_node, params, info)

    return ents | collect

def resolve_aggregate(_, info, *, type_node, **params):
    # We can potentially defer the aggregation till later, by returning a kind
    # of lazy object here. However, for simplicity in the beginning, I will
    # aggregate everything, even if the query is only for a single field.

    ents = resolve_query(_, info, type_node=type_node, filter=params.get("filter", None))

    out = {"count": len(ents)}
    for z_field in type_node > L[RT.GQL_Field] | filter[op_is_aggregable]:
        vals = ents | map[lambda z: resolve_field(z, info, z_field=z_field)] | filter[Not[equals[None]]] | collect

        field_name = z_field >> RT.Name | value | collect

        if op_is_orderable(z_field | target):
            if len(vals) == 0:
                val_min = val_max = None
            else:
                val_min = min(vals)
                val_max = max(vals)

            # if isinstance(val_min, Time):
            #     val_min = str(val_min)
            #     val_max = str(val_max)

            out[f"{field_name}Min"] = val_min
            out[f"{field_name}Max"] = val_max

        if op_is_summable(z_field | target):
            if len(vals) == 0:
                val_sum = None
                val_avg = None
            else:
                val_sum = sum(vals)
                val_avg = val_sum / len(vals)

            out[f"{field_name}Sum"] = val_sum
            out[f"{field_name}Avg"] = val_avg

    return out

def resolve_field(z, info, *, z_field, **params):
    is_list = z_field | op_is_list | collect
    is_required = z_field | op_is_required | collect

    opts = internal_resolve_field(z, info, z_field)

    if is_list:
        opts = handle_list_params(opts, target(z_field), params, info)

    opts = collect(opts)

    if is_list:
        return opts
    else:
        if not is_required and len(opts) == 0:
            return None
        return single(opts)

def resolve_id(z, info):
    return str(origin_uid(z))


##############################
# * Mutations
#----------------------------
def resolve_add(_, info, *, type_node, **params):
    # TODO: @unique checks should probably be done post change as multiple adds
    # could try changing the same thing, including nested types.
    try:
        name_gen = NameGen()
        actions = []
        post_checks = []
        new_obj_names = []
        updated_objs = []

        # This is not optimal but simplest to understand for now.
        upsert = params.get("upsert", False)
        for item in params["input"]:
            # Check id fields to see if we already have this.
            if "id" in item:
                if not upsert:
                    raise ExternalError("Can't update item with id without setting upsert")
                obj = find_existing_entity_by_id(info, type_node, item["id"])
                if obj is None:
                    raise Exception("Item doesn't exist")
                set_d = {**item}
                set_d.pop("id")
                new_actions,new_post_checks = update_entity(obj, info, type_node, set_d, {}, name_gen)
                actions += new_actions
                post_checks += new_post_checks
                updated_objs += [obj]
            else:
                obj_name,more_actions,more_post_checks = add_new_entity(info, type_node, item, name_gen)
                actions += more_actions
                post_checks += more_post_checks
                new_obj_names += [obj_name]

        r = commit_with_post_checks(actions, post_checks, info)

        ents = updated_objs + [r[name] for name in new_obj_names]
        # Note: we return the details after any updates
        ents = ents | map[now] | collect

        count = len(new_obj_names)

        return {"count": count, "ents": ents}
    except ExternalError:
        raise
    except Exception as exc:
        log.error("There was an error in resolve_add", exc_info=exc)
        raise Exception("Unexpected error")
            
def resolve_upfetch(_, info, *, type_node, **params):
    # TODO: @unique checks should probably be done post change as multiple adds
    # could try changing the same thing, including nested types.
    try:
        # Upfetch is a lot like add with upsert, except it uses the specially
        # indicated upfetch field to work on, rather than id.
        name_gen = NameGen()
        actions = []
        post_checks = []
        new_obj_names = []
        updated_objs = []

        upfetch_field = op_upfetch_field(type_node)
        field_name = upfetch_field >> RT.Name | value | collect

        for item in params["input"]:
            # Check the upfetch field to see if we already have this.
            if field_name not in item:
                raise Exception("Should never get here, because the upfetch field should be marked as required")

            obj = find_existing_entity_by_field(info, type_node, upfetch_field, item[field_name])
            if obj is None:
                obj_name,more_actions,more_post_checks  = add_new_entity(info, type_node, item, name_gen)
                actions += more_actions
                post_checks += more_post_checks
                new_obj_names += [obj_name]
            else:
                new_actions,new_post_checks = update_entity(obj, info, type_node, item, {}, name_gen)
                actions += new_actions
                post_checks += new_post_checks
                updated_objs += [obj]

        r = commit_with_post_checks(actions, post_checks, info)

        ents = updated_objs + [r[name] for name in new_obj_names]
        # Note: we return the details after any updates
        ents = ents | map[now] | collect
        count = len(new_obj_names)

        return {"count": count, "ents": ents}
    except ExternalError:
        raise
    except Exception as exc:
        log.error("There was an error in resolve_upfetch", exc_info=exc)
        raise Exception("Unexpected error")
            
        
def resolve_update(_, info, *, type_node, **params):
    # TODO: @unique checks should probably be done post change as multiple adds
    # could try changing the same thing, including nested types.
    try:
        if "input" not in params or "filter" not in params["input"]:
            raise Exception("Not allowed to update everything!")
        ents = resolve_query(_, info, type_node=type_node, filter=params["input"]["filter"])

        actions = []
        post_checks = []

        name_gen = NameGen()
        for ent in ents:
            new_actions,new_post_checks = update_entity(ent, info, type_node, params["input"].get("set", {}), params["input"].get("remove", {}), name_gen)
            actions += new_actions
            post_checks += new_post_checks

        commit_with_post_checks(actions, post_checks, info)

        count = len(ents)

        # Note: we return the details after the update
        ents = ents | map[now] | collect

        return {"count": count, "ents": ents}
    except ExternalError:
        raise
    except Exception as exc:
        log.error("There was an error in resolve_update", exc_info=exc)
        raise Exception("Unexpected error")

def resolve_delete(_, info, *, type_node, **params):
    try:
        g = info.context["g"]
        # Do the same thing as a resolve_query but delete the entities instead.
        if "filter" not in params:
            raise ExtenalError("Not allowed to delete everything!")
        ents = resolve_query(_, info, type_node=type_node, **params)

        if not ents | map[pass_delete_auth[type_node][info]] | all | collect:
            raise ExtenalError("Auth check returned False")

        [terminate[ent] for ent in ents] | transact[g] | run

        count = len(ents)

        return {"count": count, "ents": ents}
    except ExternalError:
        raise
    except Exception as exc:
        log.error("There was an error in resolve_delete", exc_info=exc)
        raise Exception("Unexpected error")

def resolve_filter_response(obj, info, *, type_node, **params):
    ents = obj["ents"]

    ents = handle_list_params(ents, type_node, params, info)

    return ents | collect




##############################################
# * Internal query parts
#--------------------------------------------

def handle_list_params(opts, z_node, params, info):
    opts = maybe_filter_result(opts, z_node, info, params.get("filter", None))
    opts = maybe_sort_result(opts, z_node, info, params.get("order", None))
    opts = maybe_paginate_result(opts, params.get("first", None), params.get("offset", None))
    return opts

@func
def field_resolver_by_name(z, type_node, info, name):
    # Note name goes last because it is curried last... this is weird
    if name == "id":
        return resolve_id(z, info=info)
    sub_field = get_field_rel_by_name(type_node, name)
    return resolve_field(z, info=info, z_field=sub_field)


# ** Filtering

def maybe_filter_result(opts, z_node, info, fil=None):
    if fil is None:
        return opts

    temp_fil = build_filter_zefop(fil, z_node, info)
    return opts | filter[build_filter_zefop(fil, z_node, info)]

def build_filter_zefop(fil, z_node, info):
    field_resolver = field_resolver_by_name[z_node][info]
    # top level is ands
    top = And
    for key,sub in fil.items():
        if key == "and":
            this = And
            for part in sub:
                this = this[build_filter_zefop(part, z_node, info)]
        elif key == "or":
            this = Or
            for part in sub:
                this = this[build_filter_zefop(part, z_node, info)]
        elif key == "not":
            this = Not[build_filter_zefop(sub, z_node, info)]
        elif key == "id":
            # This is handled specially - functions like an "in".
            val = field_resolver["id"]
            this = val | contained_in[sub]
        else:
            # This must be a field
            z_field = get_field_rel_by_name(z_node, key)
            z_field_node = target(z_field)

            val = field_resolver[key]
            if op_is_list(z_field):
                list_top = And
                for list_key,list_sub in sub.items():
                    if list_key == "any":
                        sub_fil = build_filter_zefop(list_sub, z_field_node, info)
                        list_top = list_top[val | map[sub_fil] | any]
                    elif list_key == "all":
                        sub_fil = build_filter_zefop(list_sub, z_field_node, info)
                        list_top = list_top[val | map[sub_fil] | all]
                    elif list_key == "size":
                        temp = val | length | scalar_comparison_op(list_sub)
                        list_top = list_top[temp]
                    else:
                        raise Exception(f"Unknown list filter keyword: {list_key}")
                this = list_top
            elif op_is_scalar(z_field_node):
                if isinstance(sub, bool):
                    this = val | equals[sub]
                else:
                    this = val | scalar_comparison_op(sub)
            else:
                sub_fil = build_filter_zefop(sub, z_field_node, info)
                this = val | And[Not[equals[None]]][sub_fil]


        top = top[this]

    return top

def scalar_comparison_op(sub):
    this = And[Not[equals[None]]]
    for scalar_key,scalar_sub in sub.items():
        if scalar_key == "eq":
            this = this[equals[scalar_sub]]
        elif scalar_key == "in":
            this = this[contained_in[scalar_sub]]
        elif scalar_key == "contains":
            this = this[contains[scalar_sub]]
        elif scalar_key == "le":
            this = this[less_than_or_equal[scalar_sub]]
        elif scalar_key == "lt":
            this = this[less_than[scalar_sub]]
        elif scalar_key == "ge":
            this = this[greater_than_or_equal[scalar_sub]]
        elif scalar_key == "gt":
            this = this[greater_than[scalar_sub]]
        elif scalar_key == "between":
            this = this[greater_than_or_equal[scalar_sub["min"]]]
            this = this[less_than_or_equal[scalar_sub["max"]]]
        else:
            raise Exception(f"Unknown comparison operator: {key}")
    return this
                
def get_field_rel_by_name(z_type, name):
    return (z_type > L[RT.GQL_Field]
            | filter[Z >> RT.Name | value | equals[name]]
            | first
            | collect)

# ** Sorting

def maybe_sort_result(opts, z_node, info, sort_decl=None):
    if sort_decl is None:
        return opts

    field_resolver = field_resolver_by_name[z_node][info]

    # First, get the list of things to sort by, so that we can reverse it
    sort_list = []
    cur = sort_decl
    while cur is not None:
        if "asc" in cur:
            assert not "desc" in cur
            val = field_resolver[cur["asc"]]
            sort_list += [sort[val]]
        elif "desc" in cur:
            val = field_resolver[cur["desc"]]
            sort_list += [sort[val][True]]

        cur = cur.get("then", None)

    sort_list.reverse()

    for action in sort_list:
        opts = opts | action

    return opts

# ** Pagination

def maybe_paginate_result(opts, first=None, offset=None):
    if offset is not None:
        opts = opts | skip[offset]
    if first is not None:
        opts = opts | take[first]
    return opts


# ** Resolution

@func
def internal_resolve_field(z, info, z_field):
    # This returns a LazyValue so we can deal with whatever comes out without
    # instantiating the whole list.

    # Also *only* returns that list. The calling function needs to apply the single/list logic
    
    is_incoming = z_field | op_is_incoming | collect
    # This is a delegate
    relation = z_field >> RT.GQL_Resolve_With | collect
    is_triple = source(relation) != relation

    rt = RT(relation)

    if is_triple:
        if is_incoming:
            assert rae_type(z) == rae_type(target(relation)), f"The RAET of the object {z} is not the same as that of the delegate relation {target(relation)}"
        else:
            assert rae_type(z) == rae_type(source(relation)), f"The RAET of the object {z} is not the same as that of the delegate relation {source(relation)}"

    if is_incoming:
        opts = z << L[rt]
        if is_triple:
            opts = opts | filter[is_a[rae_type(source(relation))]]
    else:
        opts = z >> L[rt]
        if is_triple:
            opts = opts | filter[is_a[rae_type(target(relation))]]

    # Auth filtering todo
    opts = opts | filter[pass_query_auth[target(z_field)][info]]

    # We must convert final objects from AEs to python types.
    # if z_field | target | is_core_scalar | collect:
    if z_field | target | op_is_scalar | collect:
        opts = opts | map[value]

    return opts


# ** Handling adding

def NameGen():
    n = 0
    while True:
        n += 1
        yield n

def find_existing_entity_by_id(info, type_node, id):
    if id is None:
        return None
    the_uid = uid(id)
    if the_uid is None:
        raise Exception("An id of {id} cannot be converted to a uid.")
    
    g = info.context["g"]

    et = ET(type_node >> RT.GQL_Delegate | collect)
    ent = g[uid(id)] | now | collect
    if not is_a(ent, et):
        return None
    if not ent | pass_query_auth[type_node][info] | collect:
        return None

    return ent

def find_existing_entity_by_field(info, type_node, z_field, val):
    if val is None:
        raise Exception("Can't find an entity by a None field value")

    g = info.context["g"]
    et = ET(type_node >> RT.GQL_Delegate | collect)

    # Note: we filter out with query auth even though this may mean we return
    # None instead of the actual entity. The follow-up logic with an upfetch
    # might seem wrong, as it would try to create the same item again. However,
    # because the field is also @unique then this will fail. Better to keep the
    # failure point at the same location so we can determine what kind of error
    # to return and whether this will leak sensitive information.
    ent = (g | now | all[et] | filter[pass_query_auth[type_node][info]]
           | filter[internal_resolve_field[info][z_field] | single_or[None] | equals[val]]
           | single_or[None]
           | collect)

    return ent

def add_new_entity(info, type_node, params, name_gen):

    actions = []
    post_checks = []

    this = str(next(name_gen))
    type_name = type_node >> RT.Name | value | collect

    post_checks += [("add", this, type_node)]

    et = ET(type_node >> RT.GQL_Delegate | collect)
    actions += [et[this]]

    # This should probably be cached
    field_mapping = {}
    for z_field in type_node > L[RT.GQL_Field]:
        field_name = value(z_field >> RT.Name)
        field_mapping[field_name] = z_field
        # Convenient spot to validate that required fields have been specified.
        if op_is_required(z_field):
            if field_name not in params:
                raise ExternalError(f"Required field '{field_name}' not given when creating new entity of type '{type_name}'")


    for key,val in params.items():
        # TODO: Validate that any unique field is not duplicated
        z_field = field_mapping[key]
        field_name = value(z_field >> RT.Name)
        rt = RT(z_field >> RT.GQL_Resolve_With | collect)
        if z_field | op_is_list | collect:
            if not isinstance(val, list):
                raise Exception(f"Value should have been list but was {type(val)}")

        if z_field | target | op_is_scalar | collect:
            if op_is_unique(z_field):
                others = info.context["g"] | now | all[et] | filter[Z >> O[rt] | value_or[None] | equals[val]] | func[set] | collect
                if len(others) > 0:
                    log.error("Trying to add a new entity with unique field that conflicts with others", et=et, field=field_name, others=others)
                    raise ExternalError(f"Unique field '{field_name}' conflicts with existing items.")
            if z_field | op_is_list | collect:
                l = val
            else:
                l = [val]

            for item in l:
                # TODO: Maybe proper ordering here too? Or should this
                # always be considered to be a set, ordered only when the
                # client requests it?
                if z_field | op_is_incoming | collect:
                    actions += [item, rt, (Z[this])]
                else:
                    actions += [(Z[this], rt, item)]
        else:
            if z_field | op_is_list | collect:
                l = val
            else:
                l = [val]

            for item in l:
                obj,obj_actions,obj_post_checks = find_or_add_entity(item, info, target(z_field), params, name_gen)
                actions += obj_actions
                post_checks += obj_post_checks
                if z_field | op_is_incoming | collect:
                    actions += [(obj, rt, Z[this])]
                else:
                    actions += [(Z[this], rt, obj)]

    return this, actions, post_checks

def find_or_add_entity(val, info, target_node, params, name_gen):
    if isinstance(val, dict) and val.get("id", None) is not None:
        obj = find_existing_entity_by_id(info, target_node, val["id"])
        return obj,[],[]
    else:
        obj_name,actions,post_checks = add_new_entity(info, target_node, val, name_gen)
        return Z[obj_name], actions, post_checks
    

def update_entity(z, info, type_node, set_d, remove_d, name_gen):
    type_name = type_node >> RT.Name | value | collect

    # Refuse to set and remove the same thing. Just too confusing to deal with
    assert(len(set(set_d.keys()).intersection(remove_d.keys())) == 0), "Can't have the same set/remove keys"
    if len(set_d) + len(remove_d) == 0:
        raise ExternalError("No set or remove in update_entity!")

    # This is the pre-update auth only
    # TODO: Post-update auth
    if not pass_pre_update_auth(z, type_node, info):
        raise ExternalError("Not allowed to update")

    actions = []
    post_checks = []
    post_checks += [("update", z, type_node)]

    # This should probably be cached
    field_mapping = {}
    for z_field in type_node > L[RT.GQL_Field]:
        field_mapping[value(z_field >> RT.Name)] = z_field

    for key,val in set_d.items():
        # TODO: Validate that any unique field is not duplicated
        z_field = field_mapping[key]
        field_name = value(z_field >> RT.Name)
        # TODO: This should be able to distinguish based on the triple, not just the RT
        rt = RT(z_field >> RT.GQL_Resolve_With | collect)

        if op_is_list(z_field):
            raise Exception(f"Updating list things is a todo (for {z_field=})")
        else:
            if op_is_unique(z_field):
                et = ET(type_node >> RT.GQL_Delegate | collect)
                others = info.context["g"] | now | all[et] | filter[Z >> O[rt] | value_or[None] | equals[val]] | collect
                others = set(others) - {z}
                if len(others) > 0:
                    log.error("Trying to modify entity with unique field that conflicts with others", z=z, et=et, field=field_name, others=others)
                    raise ExternalError(f"Unique field '{field_name}' conflicts with existing items.")

            if op_is_incoming(z_field):
                maybe_prior_rel = z < O[rt] | collect
            else:
                maybe_prior_rel = z > O[rt] | collect
            if z_field | target | op_is_scalar | collect:
                if maybe_prior_rel is None:
                    if op_is_incoming(z_field):
                        actions += [(val, rt, z)]
                    else:
                        actions += [(z, rt, val)]
                else:
                    actions += [(maybe_prior_rel | target | collect) <= val]
            else:
                raise Exception("Updating non-scalars is TODO")
    
    for key,val in remove_d.items():
        if val is not None:
            raise ExternalError("Remove vals need to be nil")
        z_field = field_mapping[key]
        field_name = value(z_field >> RT.Name)
        if op_is_required(z_field):
            raise ExternalError(f"Not allowed to remove required field '{field_name}' on type '{type_name}'")
        # TODO: This should be able to distinguish based on the triple, not just the RT
        rt = RT(z_field >> RT.GQL_Resolve_With | collect)
        if op_is_incoming(z_field):
            rels = z < L[rt]
        else:
            rels = z > L[rt]
        actions += [terminate[rel] for rel in rels]

        # Also delete scalars
        if z_field | target | op_is_scalar | collect:
            actions += [terminate[target(rel)] for rel in rels]

    return actions, post_checks


##############################
# * Auth things
#----------------------------

def pass_auth_generic(z, schema_node, info, rt_list):
    to_call = None
    for rt in rt_list:
        to_call = schema_node >> O[rt] | collect
        if to_call is not None:
            break
    else:
        return True

    # TODO: Later on these will be zef functions so easier to call
    return temporary__call_string_as_func(
        to_call | value | collect,
        z=z,
        info=info,
        type_node=schema_node
    )

@func
def pass_query_auth(z, schema_node, info):
    return pass_auth_generic(z, schema_node, info, [RT.AllowQuery])

@func
def pass_add_auth(z, schema_node, info):
    return pass_auth_generic(z, schema_node, info, [RT.AllowAdd, RT.AllowQuery])

@func
def pass_pre_update_auth(z, schema_node, info):
    return pass_auth_generic(z, schema_node, info, [RT.AllowUpdate, RT.AllowQuery])

@func
def pass_post_update_auth(z, schema_node, info):
    return pass_auth_generic(z, schema_node, info, [RT.AllowUpdatePost, RT.AllowUpdate, RT.AllowQuery])

@func
def pass_delete_auth(z, schema_node, info):
    return pass_auth_generic(z, schema_node, info, [RT.AllowDelete, RT.AllowUpdate, RT.AllowQuery])

def temporary__call_string_as_func(s, **kwds):
    from ... import core, ops
    try:
        out = eval(
            s, {
                **{name: getattr(core,name) for name in dir(core) if not name.startswith("_")},
                **{name: getattr(ops,name) for name in dir(ops) if not name.startswith("_")},
                "auth_field": P(auth_helper_auth_field, **kwds),
            },
            kwds
        )
    except Exception as exc:
        log.error("Problem calling auth expression", expr=s, exc_info=exc, kwds=kwds)
        return False

    if isinstance(out, LazyValue):
        out = collect(out)
    elif isinstance(out, ZefOp):
        out = out(kwds.get("z", None))

    return out

def auth_helper_auth_field(field_name, auth, *, z, type_node, info):
    # A helper function for graphql schema, that requests an auth check of the
    # given kind on one of its fields.

    # Only makes sense for fields that are required, such as a user field.
    try:
        z_field = get_field_rel_by_name(type_node, field_name)
        z_field_node = target(z_field)
        val = field_resolver_by_name(z, type_node, info, field_name)
    except:
        # Going to assume this is because traversal failed auth along the way somewhere.
        log.error("auth_field helper got an exception, assuming failure of auth")
        return False

    if auth == "query":
        func = pass_query_auth
    elif auth == "add":
        func = pass_add_auth
    elif auth == "update":
        func = pass_pre_update_auth
    elif auth == "updatePost":
        func = pass_post_update_auth
    elif auth == "delete":
        func = pass_delete_auth
    else:
        raise Exception(f"Don't understand auth type '{auth}' in auth_helper_auth_field")

    return func(val, z_field_node, info)



def commit_with_post_checks(actions, post_checks, info):
    g = info.context["g"]
    with Transaction(g):
        try:
            r = actions | transact[g] | run
            # Test all post checks
            for (kind,obj,type_node) in post_checks:
                if type(obj) == str:
                    obj = r[obj]
                assert type(obj) == ZefRef
                obj = now(obj)
                type_name = {type_node >> O[RT.Name] | value_or[''] | collect}

                if kind == "add":
                    if not pass_add_auth(obj, type_node, info):
                        raise Exception(f"Add auth check for type_node of '{type_name}' returned False")
                elif kind == "update":
                    if not pass_post_update_auth(obj, type_node, info):
                        raise Exception(f"Post-update auth check for type_node of '{type_name}' returned False")
        except Exception as exc:
            log.error("Aborting transaction",
                      exc_info=exc)
            from ...pyzef.internals import AbortTransaction
            AbortTransaction(g)
            raise ExternalError("Mutation did not pass auth checks")

    return r



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..core import *
from ..ops import *


# printing utils
def print_types(generated_types, cog):
    cog.out(f"object_types = [{','.join(generated_types)}]")


def print_imports(additional_exec, cog):
    cog.outl("from ariadne import ObjectType,MutationType,QueryType,InterfaceType,SubscriptionType")
    cog.outl("from zef.gql.ariadne_utils import *")
    cog.outl("from zef.gql.resolvers_utils import *")
    cog.outl("from zef import *")
    cog.outl("from zef.ops import *")
    cog.outl(additional_exec)


# general utils
def de_gqlify(s, object):
    if len(s) < 5:
        raise Exception(f"String {s} is too small to be a GQL_x string.")
    if s[:4] != "GQL_":
        raise Exception(f"String {s} does not start with GQL_.")
    s = s[4:]
    if object:
        return s
    return s[0].lower() + s[1:]


def gqlify(s):
    s = s.strip('{ ')
    return "GQL_" + s[0].upper() + s[1:]

# data utils
def parse_list_et_gql(list_et):
    aet_values = []
    arguments = list_et >> L[RT.argument] | collect
    for arg in arguments:
        if BT(arg) == BT.ENTITY_NODE:
            return ""
        else:
            aet_type = str(AET(arg | to_ezefref | collect))
            if arg | value | collect is not None:
                aet_values.append(arg | value | collect)
            else:
                return ""

    aet_values = "[" + ','.join([str(x) for x in aet_values]) + "]"
    return aet_values


# ariadne utils
def handle_params(rt, params):
    if length(rt >> L[RT(gqlify("queryParams"))]) > 0:
        for param in rt > L[RT(gqlify("queryParams"))]:
            param_name = de_gqlify(str(param >> RT.Name | value | collect), False)
            if BT(param | target | collect) == BT.ATOMIC_ENTITY_NODE:
                aet_type = str(AET(param | target | collect))
                aet_value = param | target | value | collect
                if length(param >> L[RT(gqlify("nonNullable"))]) > 0:
                    params.append(f"{param_name}")
                else:
                    params.append(f"{param_name} = {aet_value}")
            else:
                if de_gqlify(str(ET(param | target | collect)), False) == "list":
                    et_values = parse_list_et_gql(param | target | collect)
                    params.append(f"{param_name} = {et_values}")
                else:
                    params.append(param_name)
    return params


def resolve_with(rt, bt, ft):
    d_rt = rt >> RT(gqlify("resolve_with")) | collect
    d_rt_name = str(RT(d_rt))
    is_list = is_a(ft, ET(gqlify("list")))
    if is_list:
        bt = BT(ft >> RT.argument | to_ezefref | collect)
    is_out = rt > RT(gqlify("resolve_with")) >> O[RT.IsOut] | value_or[True] | collect

    dir = ">>" if is_out else "<<"

    if bt == BT.ATOMIC_ENTITY_NODE:
        if is_list:
            return f'return (z {dir} L[RT.{d_rt_name}]) | map[value] | collect'
        else:
            return f'return (z {dir} O[RT.{d_rt_name}]) | maybe_value | collect'
    else:
        if is_list:
            # return f'return (z >> L[RT.{d_rt_name}] | collect) , (z > L[RT.{d_rt_name}])| map[uid] | collect'
            return f'return (z {dir} L[RT.{d_rt_name}] | collect)'
        else:
            # return f'return (z >> RT.{d_rt_name} | collect) , (z > RT.{d_rt_name})|uid | collect'
            return f'return (z {dir} O[RT.{d_rt_name}] | collect)'


def resolve_with_script(rt):
    zef_script = rt >> RT(gqlify("resolve_with_script"))
    script = f'g["{zef_script | uid | collect}"] | now | collect'
    return f'zef_script = {script} ; return zef_execute(zef_script, z=z)'

def resolve_with_func(rt):
    z_func = rt >> RT(gqlify("resolve_with_func"))

    extra = rt >> L[RT(gqlify("queryParams"))] >> RT.Name | value | map[lambda s: de_gqlify(s,False)] | collect
    param_names = {"z", "ctx", "g", *extra}

    dict_create = '{' + ', '.join(f"'{key}': {key}" for key in param_names) + '}'
    return f'''
        resolve_func = func[g["{uid(z_func)}"]] 
        pass_dict = {dict_create}
        return pass_dict | unpack[resolve_func] | collect
    '''

def resolve_with_intermediate(rt):
    d_rt = rt >> RT(gqlify("resolve_with_intermediate"))
    d_rt_name = str(RT(d_rt))
    return f'return (z > RT.{d_rt_name}), (z > RT.{d_rt_name})|uid | collect'


def resolve_with_zef_function(rt):
    zef_function_uid = str(rt >> RT(gqlify("resolve_with_zef_function")) | uid | collect)
    params = rt > L[RT(gqlify("queryParams"))] >> RT.Name | value | map[lambda s: de_gqlify(s,False)] | collect
    defaults = ', '.join(f"{key} = {key}" for key in {"z", "ctx", "g"})
    params = ', '.join(i for i in params) + "," if len(params) != 0 else ""
    return f'return g["{zef_function_uid}"]({params}{defaults} )'


def resolve_with_wrapper(fn_body):
    return f"""
    try:
        {fn_body}
    """ + """
    except Exception as e:
        if "error traversing" not in str(e):
            log.warn(f"Handled exception while calling a resolve_with_*", exc_info=e)
        return None
    """


def create_function_body(ot, ft, bt, rt, fn, customer_specific_resolvers):
    default_params = ["z", "ctx"]
    params = handle_params(rt, default_params)
    # TODO (Cleanup) might consider making these as well customer specific
    # Resolve_With relations have highest priority
    if length(rt >> L[RT(gqlify("resolve_with"))]) > 0:
        return resolve_with_wrapper(resolve_with(rt, bt, ft)), params

    elif length(rt >> L[RT(gqlify("resolve_with_script"))]) > 0:
        return resolve_with_wrapper(resolve_with_script(rt)), params

    elif length(rt >> L[RT(gqlify("resolve_with_zef_function"))]) > 0:
        return resolve_with_wrapper(resolve_with_zef_function(rt)), params

    elif length(rt >> L[RT(gqlify("resolve_with_func"))]) > 0:
        return resolve_with_wrapper(resolve_with_func(rt)), params

    elif length(rt >> L[RT(gqlify("resolve_with_intermediate"))]) > 0:
        return resolve_with_wrapper(resolve_with_intermediate(rt)), params

    elif length(rt >> L[RT(gqlify("resolve_with_body"))]) > 0:
        # return (rt >> RT(gqlify("resolve_with_body")) | value | collect), params
        return resolve_with_wrapper(rt >> RT(gqlify("resolve_with_body")) | value | collect), params

    return customer_specific_resolvers(ot, ft, bt, rt, fn)

def initialize_object_type(object_type, cog):
    if "Mutation" == object_type:
        cog.outl(f'{object_type} = MutationType()')
    elif "Subscription" == object_type:
        cog.outl(f'{object_type} = SubscriptionType()')
    else:
        cog.outl(f'{object_type} = ObjectType("{object_type}")')


def generate_object_resolver(object_type, field_name, rt, fn_body, params, cog):
    def custom_sort(item):
        if "=" in item: return 1
        return -1
    params = params | sort[custom_sort] | collect
    cog.outl(f'@{object_type}.field("{field_name}")')
    cog.outl(f'def resolve_{object_type}_{uid(rt)}({", ".join(params)}):')
    cog.outl(f'     {fn_body}\n')


def generate_enum_resolver(e, cog):
    enum_name = de_gqlify(str(e >> RT.Name | value | collect), True)
    cog.outl(f'{enum_name} = EnumType("{enum_name}", {"{"}')
    i = -1
    for f in (e >> L[RT(gqlify("field"))]):
        i += 1
        en_value = (f | value).enum_value

        cog.outl(f' "{en_value}": {i},')
    cog.outl(f'{"}"},)\n')
    return enum_name


def generate_interface_resolver(i, cog):
    interface_name = de_gqlify(str(i >> RT.Name | value | collect), True)
    cog.outl(f'{interface_name} = InterfaceType("{interface_name}")')
    cog.outl(f'@{interface_name}.type_resolver')
    cog.outl(f'def resolve_{interface_name}_type_resolver(obj, *_):')
    cog.outl(f'#TODO make sure that those names are consistent with your Data Graph type name')
    for t in (i << L[RT(gqlify("implements"))]):
        object_name = de_gqlify(str(t >> RT.Name | value | collect), True)
        cog.outl(f'    if str(ET(obj)) == "{object_name}":')
        cog.outl(f'        return "{object_name}"')
    cog.outl(f'    if str(ET(obj)) == "Step":')
    cog.outl(f'        return "ProcessOrder"')
    cog.outl(f'    return None')
    return interface_name


def generate_all(global_dict, cog):
    root = global_dict["schema_root"]
    default_resolvers_list = global_dict["default_resolvers_list"]
    customer_specific_resolvers = global_dict["customer_specific_resolvers"]
    object_types = []
    types = root >> L[RT(gqlify("type"))]
    interfaces = root >> L[RT(gqlify("interface"))]

    for t in types:
        object_type = de_gqlify(str(t >> RT.Name | value | collect), True)
        # Don't generate resolvers for function in this list
        if object_type in default_resolvers_list:
            continue
        object_types.append(object_type)
        initialize_object_type(object_type, cog)

        for rt in (t > L[RT(gqlify("field"))]):
            field_name = de_gqlify(str(rt >> RT.Name | value | collect), False)
            field_type = (rt | target | to_ezefref | collect)
            fn_body, params = create_function_body(object_type, field_type, BT(field_type), rt, field_name, customer_specific_resolvers)
            generate_object_resolver(object_type, field_name, rt, fn_body, params, cog)

    for i in interfaces:
        object_types.append(generate_interface_resolver(i, cog))

    return object_types



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from operator import ne
from ._ops import *
from dataclasses import dataclass
from . import VT


@dataclass
class Val:
    arg: VT.Any

class FlatGraph:
    def __init__(self, *args):
        if args == ():
            self.key_dict = {}
            self.blobs = ()
        elif len(args) == 1 and isinstance(args[0], list):
            new_fg = FlatGraph()
            new_fg = new_fg | insert[args[0]] | collect
            self.key_dict = new_fg.key_dict
            self.blobs = new_fg.blobs
        else:
            raise NotImplementedError("FlatGraph with args")

    def __repr__(self):
        kdict = "\n".join([f"({k}=>{v})" for k,v in self.key_dict.items()])
        blobs = "\n".join([str(e) for e in self.blobs])
        return f'FlatGraph(\n{kdict}\n-------\n{blobs}\n)'
    
    def __or__(self, other):
        return LazyValue(self) | other

    def __getitem__(self, key):
        return get(self, key)


class FlatRef:
    def __init__(self, fg, idx):
        self.fg = fg
        self.idx = idx
    def __repr__(self):
        return f'<FlatRef #{abs(self.idx)} {repr(self.fg.blobs[self.idx][1])}>'
    
    def __or__(self, other):
        return LazyValue(self) | other

    def __gt__(self, other):
        return LazyValue(self) > other

    def __lt__(self, other):
        return LazyValue(self) < other

    def __lshift__(self, other):
        return LazyValue(self) << other
    
    def __rshift__(self, other):
        return LazyValue(self) >> other

class FlatRefs:
    def __init__(self, fg, idxs):
        self.fg = fg
        self.idxs = idxs
    def __repr__(self):
        newline = '\n'
        return f"""<FlatRefs len={len(self.idxs)}> [
{["    " + repr(FlatRef(self.fg, i)) for i in self.idxs] | join[newline] | collect}
]"""
    
    def __or__(self, other):
        return LazyValue(self) | other

    def __iter__(self):
        return (FlatRef(self.fg, i) for i in self.idxs)

    def __gt__(self, other):
        return LazyValue(self) > other

    def __lt__(self, other):
        return LazyValue(self) < other

    def __lshift__(self, other):
        return LazyValue(self) << other
    
    def __rshift__(self, other):
        return LazyValue(self) >> other


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sdl2 import *
import ctypes
import OpenGL.GL as gl

import imgui
from imgui.integrations.sdl2 import SDL2Renderer

def main(window_name, init_state, render_func, *args, **kwds):
    window, gl_context = impl_pysdl2_init(window_name)
    imgui.create_context()
    impl = SDL2Renderer(window)

    running = True
    event = SDL_Event()

    state = init_state(*args, **kwds)
    while running:
        while SDL_PollEvent(ctypes.byref(event)) != 0:
            if event.type == SDL_QUIT:
                running = False
                break
            impl.process_event(event)
        impl.process_inputs()

        imgui.new_frame()

        render_func(state)

        gl.glClearColor(1., 1., 1., 1)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT)

        imgui.render()
        impl.render(imgui.get_draw_data())
        SDL_GL_SwapWindow(window)

    impl.shutdown()
    SDL_GL_DeleteContext(gl_context)
    SDL_DestroyWindow(window)
    SDL_Quit()


def impl_pysdl2_init(window_name):
    width, height = 1280, 720

    if SDL_Init(SDL_INIT_EVERYTHING) < 0:
        print("Error: SDL could not initialize! SDL Error: " + SDL_GetError().decode("utf-8"))
        exit(1)

    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1)
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24)
    SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8)
    SDL_GL_SetAttribute(SDL_GL_ACCELERATED_VISUAL, 1)
    SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 1)
    SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, 16)
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_FLAGS, SDL_GL_CONTEXT_FORWARD_COMPATIBLE_FLAG)
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 4)
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 1)
    SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_CORE)

    SDL_SetHint(SDL_HINT_MAC_CTRL_CLICK_EMULATE_RIGHT_CLICK, b"1")
    SDL_SetHint(SDL_HINT_VIDEO_HIGHDPI_DISABLED, b"1")

    window = SDL_CreateWindow(window_name.encode('utf-8'),
                              SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED,
                              width, height,
                              SDL_WINDOW_OPENGL|SDL_WINDOW_RESIZABLE)

    if window is None:
        print("Error: Window could not be created! SDL Error: " + SDL_GetError().decode("utf-8"))
        exit(1)

    gl_context = SDL_GL_CreateContext(window)
    if gl_context is None:
        print("Error: Cannot create OpenGL Context! SDL Error: " + SDL_GetError().decode("utf-8"))
        exit(1)

    SDL_GL_MakeCurrent(window, gl_context)
    if SDL_GL_SetSwapInterval(1) < 0:
        print("Warning: Unable to set VSync! SDL Error: " + SDL_GetError().decode("utf-8"))
        exit(1)

    return window, gl_context




# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *
import zef

class MyTestCase(unittest.TestCase):
    def test_zefrefs_from_slices(self):
        g = Graph()

        g_clones = []
        for i in range(10):
            g_clones += [zef.pyzef.internals.create_partial_graph(g, g.graph_data.write_head)]
            [ET.Machine]*10 | g | run

        # Add in a couple of manual edge cases
        g_clones += [zef.pyzef.internals.create_partial_graph(g, g.graph_data.write_head)]
        z = ET.Machine | g | run
        g_clones += [zef.pyzef.internals.create_partial_graph(g, g.graph_data.write_head)]
        [(z, RT.Something, z), (z, RT.Something2, z)] | g | run
        g_clones += [zef.pyzef.internals.create_partial_graph(g, g.graph_data.write_head)]
        z | terminate | g | run
        g_clones += [g]

        for i,tx in enumerate(g | all[TX] | drop[1] | collect):
            g_clone_before_tx = g_clones[i]

            for j in range(i+1,len(g_clones)):
                g_partial = zef.pyzef.internals.create_partial_graph(g_clones[j], g_clone_before_tx.graph_data.write_head)
                self.assertEqual(g_clone_before_tx.hash(), g_partial.hash())

        

if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#%%
""" 
Until we have flat graphs, we represent zefops as tuples of tuples.
The outer dimensio represents the data flow direction.



General Structure:

**************************** LazyValues ****************************
- what we actually mean: an iterable
- pull based, potentially lazy
- value based semantics
- can be sent over the wire
- representation: (initial_val, ZefOp)
- storage in evaluated form simply means that ZefOp = None
- only compute values when 
    a) piped through 'collect'      
    b) the user casts to some data type, e.g. int(my_lazy_value)  will trigger compute
    
    
**************************** Awaitables ****************************
- what we mean: an observable
- push based
- does not hold on to / cache results (do we want to enable that functionality?)
- Is it lazy? 
    a) yes, in the sense that cmpute is only performed if someone is listening
    b) but it then performs compute as soon as a new value comes in
    c) compared to iterables where there is internal freedom to transform 
       the data stage-wise or element-wise



***** should zefops be able to deal with switching between types within one operator chain? *****
unsure about all possible combinations one may want. Let's aggregate examples

(
my_lazy_value 
| filter[...] 
| map[...] 
| first         # let us assume this returns a Awaitable()
) | map[...] | subscribe[...]

Claim: 
We would ever only do anything if there is an evaluating keyword at the very end:
subscribe / for_each / collect.
Therefore we don't ever have to try to infer intermediate types. Just keep a look out 
for evaluating keywords.






********************************* Evaluating Operators *********************************
Side effects: 
subscribe / for_each

Pure value: 
collect
Further chaining possible.


********************************* Expect Operator *********************************
Introduce the 'expect' operator?
In C++ one may consider making it templated.
my_zefop | expect[Int] | ...
Will throw if during evaluation a value of a different type passes through.
Can be used as an important tool by the runtime operator introspector, e.g.
if no types can be inferred due to missing type annotations in functions.

Op
RT
LazyValue
Awaitable
CollectingOp 
ForEachingOp 
SubscribingOp


LazyValue | Op -> LazyValue
Awaitable | Op -> Awaitable
Op | Op -> Op

LazyValue | CollectingOp -> LazyValue                           # The lazy_value is the general evaluated 'value' type
LazyValue | ForEachingOp -> Disposable
LazyValue | SubscribingOp -> DisposableProcessHandle        # the op should convert to an observable at some point

Awaitable | CollectingOp -> ???                            # we could allow Awaitable | CollectingOp -> Buffer
Awaitable | ForEachingOp -> ???
Awaitable | SubscribingOp -> RunningProcessHandle

Op | CollectingOp  -> CollectingOp
Op | ForEachingOp  -> ForEachingOp
Op | SubscribingOp -> SubscribingOp

CollectingOp | Op -> Op                                 # the 'collect' can be inserted as an elementary op
SubscribingOp | Op -> ???
ForEachingOp | Op -> ???

ZefRef.AET | Op -> Awaitable                               # should the Op be required to start with 'to_Awaitable()'?
RXObservable | Op -> Awaitable
ZefRef.AET | SubscribingOp -> RunningProcessHandle
RXObservable | SubscribingOp -> RunningProcessHandle


# we don't need a separate Collect, ForEach, Subscribe: these are just minimal cases of CollectingOp, ...


#************************************ Imperative Shell Interface for Awaitables: how to manually push data in ******************************************
Use the same Awaitable object type, but init with 
s1 = Awaitable(pushable=True)
s1.on_next('test event')
???????????? how to deal with values pushed in before any observer is subscribed ??????????????????



****************************** iterating over vs collecting an lazy_value **********************************
"for el in my_lazy_value"        # we don't need to evaluate all. Calculate lazily. Return wrapped itertools for now
my_lazy_value | collect          # we are free to also choose the same lazy eval up to completion. Or do 'breadth first' internally.


my_op1 = (
 expect[Awaitable[Dict]]
 | map[my_f]
 | merge
 | group_by[get_value['type']]
 | merge 
)

my_subscribing_op = my_op1 | subscribe[side_eff_fct]





*********************************** How do we want to declare new zefops? ********************************

map = ZefOp(
    RT.Map,                                     # data type to represent this op
    (),                                         # curried in params for the raw op
    get_item_handler = map_get_item_handler,    # custom behavior to deal with this_op[x]
    repr = map_repr,                            # how shoud this be displayed?
    apply_to_lazy_value_fct = ...
    apply_to_Awaitable()_fct = ...
    # verification fcts?
)




**** Composite ZefOps ****

my_op = map[f1] | filter[f2]

"""

# TODO fix repr for L op
# TODO make repr for ZefOp better
import warnings
from inspect import isfunction, getfullargspec
from types import LambdaType
from typing import Generator, Iterable, Iterator
from ._core import *
from . import internals, VT
from .internals import BaseUID, EternalUID, ZefRefUID
from ..pyzef import zefops as pyzefops
from abc import ABC

class TraversableABC(ABC):
    pass
for t in [RelationType, internals.BlobType]:
    TraversableABC.register(t)

# Anything that can appear on the right of a pipe in an op chain: x | y | z
class OpLike(ABC):
    pass


# this is used to circumvent the python '<' and '>' operator resolution rules
_terrible_global_state = {}
_call_0_args_translation = {
    RT.Now : pyzefops.now,
}
_call_n_args_translation = {}
_sources = {}

def unpack_ops(rt, ops):
    if len(ops) > 1:
        return ((rt, ops[0]),  *ops[1:])
    else:
        return ((rt, *ops),)

def unpack_tmpzefop(rt, ops):
    return ((rt, (ops[0][1])), *ops[1:])

def is_tmpzefop(arg):
    return arg.el_ops[0][0] == RT.TmpZefOp

def is_valid_LorO_op(other) -> bool:
    return (
        isinstance(other, ZefOp) 
        and len(other.el_ops)==1 
        and other.el_ops[0][0] in [RT.L, RT.O]
        )

def is_supported_stream(o):
    from rx.subject import Subject
    from rx.core import Observable
    if type(o) in {Subject, Observable}: return True
    return False

def is_python_scalar_type(o):
    return type(o) in {str, bytes, int, bool, float, type(None)}

def is_supported_value(o):
    from .fx.fx_types import _Effect_Class
    from .VT import ValueType_
    from . import  GraphSlice
    from types import GeneratorType
    from .. import Image
    from .error import _ErrorType
    from ..pyzef.main import Keyword
    from ..core.bytes import Bytes_
    if is_python_scalar_type(o): return True
    if type(o) in {set, range, GeneratorType, list, tuple, dict, _Effect_Class, ValueType_, GraphSlice, Time, Image, Bytes_, _ErrorType, Keyword}: return True
    return False

def is_supported_zef_value(o):
    from .abstract_raes import Entity, Relation, AtomicEntity
    if type(o) in {ZefRef, ZefRefs, ZefRefss, EZefRef, EZefRefs, EZefRefss, Graph, BaseUID, EternalUID, ZefRefUID, QuantityFloat, QuantityInt, Entity, Relation, AtomicEntity, Delegate, EntityType, RelationType, AtomicEntityType}: return True
    return False

def is_supported_on_subscription(o, op):
    return op.el_ops[0][0] == RT.On and (is_supported_zef_value(o) or isinstance(o, tuple))

def op_chain_pretty_print(el_ops):
    if not isinstance(el_ops, list) and not isinstance(el_ops, tuple):
        return repr(el_ops)

    from ._ops import map, collect, to_snake_case
    def param_to_str(pp):
        return f"[{repr(pp)}]"

    def el_op_to_str(p):
        # if p[0] == RT.OutOutOld:
        #     return f"\n>> todo!!!!"            
        return to_snake_case(str(p[0])) + ''.join([param_to_str(pp) for pp in p[1]])
    return ' | '.join(el_ops | map[el_op_to_str] | collect)

#   _                          ___                  ___                    _                                _           _    _               
#  | |      __ _  ____ _   _  / _ \  _ __   ___    |_ _| _ __ ___   _ __  | |  ___  _ __ ___    ___  _ __  | |_   __ _ | |_ (_)  ___   _ __  
#  | |     / _` ||_  /| | | || | | || '_ \ / __|    | | | '_ ` _ \ | '_ \ | | / _ \| '_ ` _ \  / _ \| '_ \ | __| / _` || __|| | / _ \ | '_ \ 
#  | |___ | (_| | / / | |_| || |_| || |_) |\__ \    | | | | | | | || |_) || ||  __/| | | | | ||  __/| | | || |_ | (_| || |_ | || (_) || | | |
#  |_____| \__,_|/___| \__, | \___/ | .__/ |___/   |___||_| |_| |_|| .__/ |_| \___||_| |_| |_| \___||_| |_| \__| \__,_| \__||_| \___/ |_| |_|
#                      |___/        |_|                            |_|                                                                       

class Evaluating:
    def __repr__(self):
        return "evaluating"
evaluating = Evaluating()

class ZefOp:    
    def __init__(self, el_ops: tuple):
        if isinstance(el_ops, tuple):
            self.el_ops = el_ops
        else:
            self.el_ops = (el_ops, )
    
    def __len__(self):
         return len(self.el_ops)
        
    def __repr__(self):
        return op_chain_pretty_print(self.el_ops)
    
    def __le__(self, other):
        # we want to allow for "AET.Float['some_name'] <= 42.1"
        # We convert the first expression to a "instantiated[AET.Float]['some_name']"
        from ._ops import assign_value
        return LazyValue(self) | assign_value[other]
    
    def __ror__(self, other):
        if isinstance(other, TraversableABC):
            return ZefOp(((RT.TmpZefOp, (other,)), )) | self
        elif isinstance(other, CollectingOp) or isinstance(other, ForEachingOp)  or isinstance(other, SubscribingOp):
            return other.__ror__(self)
        elif is_supported_on_subscription(other, self):
            from .op_implementations.implementation_typing_functions import on_implementation
            return on_implementation(other, *self.el_ops[0][1])
        elif is_supported_stream(other):
            from .fx import FX, Effect
            from ._ops import run
            stream =  Effect({'type': FX.Stream.CreatePushableStream}) | run
            return stream | self
        elif is_supported_value(other) or is_supported_zef_value(other):
            return LazyValue(other) | self
        # This is just for a bit of help for users to understand what's going on
        elif isinstance(other, Iterable):
            raise TypeError("An arbitrary iterable will not be automatically converted to a LazyValue when being piped into a ZefOp. This is because it cannot be determined if it is a stateful iterator (e.g. a generator or a file object) or a stateless iterable (e.g. a list or a range() object).\n\nIf you know your iterable is immutable, then wrap the iterable in `LazyValue`, that is use `LazyValue(itr) | zefop` instead of `itr | zefop`.\n\nIf your iterable is an iterator that mutates on iteration you should either\n\ta) first wrap it in a caching mechanism (TODO make this available in zef: not available in any external library that I can see) or\n\tb) make its content available as a Stream, e.g. TODO.")

        return NotImplemented
        

    def __or__(self, other):
        if isinstance(other, ZefOp):
            res = ZefOp( (*self.el_ops, *other.el_ops) )
            return res
        return NotImplemented
        

    def __lshift__(self, other):
        if isinstance(other, TraversableABC):
            return ZefOp( (*self.el_ops, (RT.InInOld, (other, ))) )
        if is_valid_LorO_op(other):
            return ZefOp( (*self.el_ops, *unpack_ops(RT.InInOld, other.el_ops)))
        if is_tmpzefop(other):
            return ZefOp( (*self.el_ops, *unpack_tmpzefop(RT.InInOld, other.el_ops))) 
        raise TypeError(f'Unexpected type in "ZefOp << ..." expression. Only an RT or L[RT] makes sense here. It was called with {other} of type {type(other)}' )
        

    def __rshift__(self, other):
        if isinstance(other, TraversableABC):
            return ZefOp( (*self.el_ops, (RT.OutOutOld, (other, ))) )
        if is_valid_LorO_op(other):
            return ZefOp( (*self.el_ops, *unpack_ops(RT.OutOutOld, other.el_ops)))
        if is_tmpzefop(other):
            return ZefOp( (*self.el_ops, *unpack_tmpzefop(RT.OutOutOld, other.el_ops)))
        raise TypeError(f'Unexpected type in "ZefOp >> ..." expression. Only an RT or L[RT] makes sense here. It was called with {other} of type {type(other)}' )
    
    def __rrshift__(self, other):
        if is_supported_zef_value(other):
            return LazyValue(other) >> self
        raise TypeError(f'Unexpected type in "ZefType >> ZefOp" expression. It was called with {other} of type {type(other)}' )

    def __rlshift__(self, other):
        if is_supported_zef_value(other):
            return LazyValue(other) << self
        raise TypeError(f'Unexpected type in "ZefType << ZefOp" expression. It was called with {other} of type {type(other)}' )
    
    def __eq__(self, other):
        if isinstance(other, ZefOp):
            return self.el_ops == other.el_ops
        return NotImplemented

    def __hash__(self):
        return hash(self.el_ops)

    def lt_gt_behavior(self, other, rt):
        # other could be one of 3 cases
        # 1) It is another ZefOp
        # 2) It is another ZefOp that is of instance TmpZefOp
        # 3) It is of type RelationType
        import inspect
        frame = inspect.currentframe().f_back.f_back
        line_no = frame.f_lineno        #identify reuse within expression from line_no: even when splitting lines for one expression, the first line is used throughout
        other_id = None
        # Start by checking if self has been previously evaluated
        if (line_no, id(self.el_ops)) in _terrible_global_state:
            cached = _terrible_global_state[(line_no, id(self.el_ops))]
        elif (line_no, id(self)) in _terrible_global_state: # If it is an RT
            cached = _terrible_global_state[(line_no, id(self))]
        else:
            cached = self.el_ops

        if isinstance(cached, LazyValue):
            base = (*cached.el_ops.el_ops,)
        else:
            base = (*(cached),)

        if isinstance(other, TraversableABC):
            res = ZefOp( (*base, (rt, (other, ))) )
            other_id = id(other)
        elif isinstance(other, ForEachingOp) or isinstance(other, SubscribingOp) or isinstance(other, CollectingOp):
            if len(other.el_ops) > 0 and is_tmpzefop(other):
                res = type(other)(ZefOp((*base, *unpack_tmpzefop(rt, other.el_ops))))
            else:
                res = type(other)(ZefOp((*base, *unpack_ops(rt, other.el_ops))))
            if type(other) in {SubscribingOp, ForEachingOp}: 
                res.func = other.func
                other.func = None

        elif is_tmpzefop(other):
            res = ZefOp( (*base, *unpack_tmpzefop(rt, other.el_ops)))
        else:
            res = ZefOp( (*base, *unpack_ops(rt, other.el_ops)))
        
        if isinstance(cached, LazyValue):
            new_lv = LazyValue(cached.initial_val)
            new_lv.el_ops = res
            res = new_lv
            if should_trigger_eval(res): return evaluate_lazy_value(res)
        
        other_id = other_id or id(other.el_ops) # Before checking the types above we can't assume other is a ZefOp
        _terrible_global_state[(line_no, other_id)] = res.el_ops
        return res

    def __gt__(self, other):
        if is_supported_zef_value(other): return LazyValue(other) < self
        return self.lt_gt_behavior(other, RT.OutOld)

    def __lt__(self, other):
        if is_supported_zef_value(other): return LazyValue(other) > self
        return self.lt_gt_behavior(other, RT.InOld)
        
    def __getitem__(self, x):
        if not len(self.el_ops)==1: 
            raise Exception(f'You can only curry arguments / parameters into an elementary zefop, i.e. one that is not chained. It was attempted to curry [{self.el_ops}] into {self}')

        return ZefOp(((self.el_ops[0][0], (*self.el_ops[0][1], x)), ))

    def __iter__(self):
        return (ZefOp((op,)) for op in self.el_ops)


    def __call__(self, *args, **kwargs):

        if len(kwargs) > 0 and self.el_ops[0][0] == RT.Function:
            from .op_implementations.dispatch_dictionary import _op_to_functions
            if len(args) > 1: extra = args[1:]
            else: extra = []
            return _op_to_functions[self.el_ops[0][0]][0](args[0], *self.el_ops[0][1], *extra, **kwargs)

        # now()
        if len(self.el_ops) == 1 and len(args) == 0:
            if self.el_ops[0][0] in  _call_0_args_translation: 
                return _call_0_args_translation[self.el_ops[0][0]]()
            raise NotImplementedError(f"A ZefOp with 0 args was called but a 0-arg dispatch doesn't exist! ({self.el_ops[0][0]!r})")
        
        # a = map[...] | last ; a() 
        if len(self.el_ops) > 1 and len(args) == 0:
            raise NotImplementedError("A ZefOp with multiple internal ops and 0 args was called!")      
            
        # run(42 | f1 | print)
        if len(self.el_ops) == 1 and is_evaluating_run(self):
            res = args[0] | self
            return res
        # x = map[...] | last ; x([1,2,3])
        # x = [1,2,3] | map[...] ; last(x)
        if len(args) == 1: 
            collect_op = CollectingOp(self) 
        else:
            curried_op = self
            for arg in args[1:]:
                curried_op = curried_op[arg]
            collect_op = CollectingOp(curried_op)
        
        lzy_val = LazyValue(args[0]) if not isinstance(args[0], LazyValue) else args[0]
        res = lzy_val | collect_op
        # Raise if this didn't evaluate!
        if isinstance(res, CollectingOp):
            raise Exception(f"ZefOp call didn't evaluate! {res}")
        return res
OpLike.register(ZefOp)

class CollectingOp:
    def __init__(self, other: ZefOp):
        self.el_ops = other.el_ops

    def __repr__(self):
        return f"CollectingOp({op_chain_pretty_print(self.el_ops)})"
        
    def __ror__(self, other):
        if isinstance(other, ZefOp): 
            return CollectingOp(ZefOp((*other.el_ops, *self.el_ops)))
        if isinstance(other, LazyValue):
            other.el_ops = CollectingOp(ZefOp((*other.el_ops, *self.el_ops)))
            return evaluate_lazy_value(other)
        if isinstance(other, TraversableABC):
            return CollectingOp(ZefOp(((RT.TmpZefOp, (other,)), *self.el_ops)))
        if is_supported_value(other) or is_supported_zef_value(other):
            return LazyValue(other) | self
        raise TypeError(f'We should not have landed here. Value passed {other} of type {type(other)}.')

    def __or__(self, other):
        base = (*self.el_ops, (RT.Collect, ()))
        if isinstance(other, ForEachingOp) or isinstance(other, SubscribingOp):
            res = type(other)(ZefOp((*base, *other.el_ops)))
            res.func = other.func
            other.func = None
            return res
        elif isinstance(other, ZefOp):
            return ZefOp((*base, *other.el_ops))

        return NotImplemented

    def lt_gt_behavior(self, other, rt):
        import inspect
        frame = inspect.currentframe().f_back.f_back
        line_no = frame.f_lineno    

        # We found a cached CollectOp so we will use it
        if (line_no, id(self)) in _terrible_global_state:
            ops = _terrible_global_state[(line_no, id(self))].el_ops
        else:
            ops = self.el_ops
        base = ((RT.Collect, ()),) if ops == () else (*ops, (RT.Collect, ()))
        
        if isinstance(other, TraversableABC):
            res = ZefOp((*base, (rt, (other,))))
        elif isinstance(other, ForEachingOp) or isinstance(other, SubscribingOp):
            if len(other.el_ops) > 0 and other.el_ops[0][0] == RT.TmpZefOp:
                res = type(other)(ZefOp((*base, (rt, (other.el_ops[0][1])), *other.el_ops[1:])))
            else:
                res = type(other)(ZefOp((*base, *unpack_ops(rt, other.el_ops))))
            res.func = other.func
            other.func = None
        elif isinstance(other,ZefOp) and is_tmpzefop(other):
            res = ZefOp( (*base, *unpack_tmpzefop(rt, other.el_ops)))
        else:
            res = ZefOp((*base, *unpack_ops(rt, other.el_ops)))

        _terrible_global_state[(line_no, id(other))] = res.el_ops
        return res

    def __gt__(self, other):
        if is_supported_zef_value(other): return LazyValue(other) < self
        return self.lt_gt_behavior(other, RT.OutOld)

    def __lt__(self, other):
        if is_supported_zef_value(other): return LazyValue(other) > self
        return self.lt_gt_behavior(other, RT.InOld)

    def lshift_rshift_behavior(self, other, rt):
        base = ((RT.Collect, ()),) if self.el_ops == () else (*self.el_ops, (RT.Collect, ()))
        if isinstance(other, TraversableABC):
            res = ZefOp((*base, (rt, (other,))))
        elif isinstance(other,ZefOp) and other.el_ops[0][0] == RT.TmpZefOp:
            res = ZefOp((*base, *unpack_tmpzefop(rt, other.el_ops)))
        else:
            res = ZefOp((*base, *unpack_ops(rt, other.el_ops)))
        return res

    def __lshift__(self, other):
        return self.lshift_rshift_behavior(other, RT.InInOld)

    def __rshift__(self, other):
        return self.lshift_rshift_behavior(other, RT.OutOutOld)

    def __rrshift__(self, other):
        if is_supported_zef_value(other):
            return LazyValue(other) >> self
        raise TypeError(f'Unexpected type in "ZefType >> CollectingOp" expression. It was called with {other} of type {type(other)}' )

    def __rlshift__(self, other):
        if is_supported_zef_value(other):
            return LazyValue(other) << self
        raise TypeError(f'Unexpected type in "ZefType << CollectingOp" expression. It was called with {other} of type {type(other)}' )
        
    def __eq__(self, other):
        return self.el_ops == other.el_ops

    def __call__(self, *args):
        # collect()
        if len(args) == 0:
            raise Exception(f"Cannot call collect() without any args!")
            
        # collect([1,2,3] | map[...] | last)
        if len(args) == 1:
            return args[0] | self
OpLike.register(CollectingOp)


class ConcreteAwaitable:
    def __init__(self, concrete_awaitable, concrete_type, chain):
        self.concrete_awaitable = concrete_awaitable
        self.concrete_type = concrete_type
        self.chain = chain + [self.concrete_type] 

    def __repr__(self):
        return f"{self.concrete_awaitable} -> {self.chain}"

class Awaitable:
    
    def __init__(self, stream_ezefref, pushable=False, unwrapping = False):
        self.stream_ezefref = stream_ezefref
        self.pushable = pushable
        self.unwrapping = unwrapping
        self.el_ops = ZefOp(())

    def __repr__(self):
        return f"Awaitable({self.stream_ezefref} \n| {self.el_ops})"

    def __or__(self, other):
        if isinstance(other, ZefOp) and is_evaluating_run(other):
            return self.evaluation(other, "Run")

        if isinstance(other, ZefOp):
            new_awaitable = Awaitable(self.stream_ezefref, self.pushable, self.unwrapping)
            if len(self.el_ops) > 0:
                new_awaitable.el_ops = ZefOp((*self.el_ops.el_ops, *other.el_ops,)) 
            else:
                new_awaitable.el_ops = other
            return new_awaitable

        if isinstance(other, SubscribingOp):
            return self.evaluation(other, "Subscribe")

        if isinstance(other, CollectingOp):
            raise NotImplementedError(f"Awaitable | not implemented with collect")

        raise NotImplementedError(f"Awaitable | not implemented with {type(other)}")
        

    def lt_gt_lshift_rshift_behavior(self, other, rt):
        import inspect
        frame = inspect.currentframe().f_back.f_back
        line_no = frame.f_lineno   
        if len(self.el_ops) > 0:
            if rt == RT.OutOld:
                res = self.el_ops > other
            elif rt == RT.InOld:
                res = self.el_ops < other
            elif rt == RT.InInOld:
                res = self.el_ops << other
            elif rt == RT.OutOutOld:
                res = self.el_ops >> other
        else:
            if isinstance(other,ZefOp) and is_tmpzefop(other):
                res = ZefOp((*unpack_tmpzefop(rt, other.el_ops),))
            elif isinstance(other, CollectingOp) or isinstance(other, SubscribingOp):
                if is_tmpzefop(other):
                    res = type(other)(ZefOp((*unpack_tmpzefop(rt, other.el_ops), )))
                else:
                    res = type(other)(ZefOp((*unpack_ops(rt, other.el_ops),))) 

                if isinstance(other, SubscribingOp): res.func = other.func

            elif isinstance(other, TraversableABC):
                res = ZefOp(((rt, (other,)), ))
            else:
                res = ZefOp((*unpack_ops(rt, other.el_ops),))
        
        new_awaitable = Awaitable(self.stream_ezefref, self.pushable, self.unwrapping)
        new_awaitable.el_ops = res

        if rt == RT.OutOld or rt == RT.InOld:  _terrible_global_state[(line_no, id(other))] = new_awaitable
        return new_awaitable

    def __gt__(self, other):
        return self.lt_gt_lshift_rshift_behavior(other, RT.OutOld)

    def __lt__(self, other):
        return self.lt_gt_lshift_rshift_behavior(other, RT.InOld)

    def __lshift__(self, other):
        return self.lt_gt_lshift_rshift_behavior(other, RT.InInOld) 

    def __rshift__(self, other):
        return self.lt_gt_lshift_rshift_behavior(other, RT.OutOutOld) 

    def evaluation(self, other, kind):
        import rx
        from rx import operators as rxops
        from rx.subject import Subject
        from rx.core import Observable
        from .op_implementations.dispatch_dictionary import _op_to_functions
        from .fx import _state
        from ._ops import absorbed, only

        curr_type =  VT.Awaitable[VT.List[VT.String]]
        source = _state['streams'][self.stream_ezefref]  
        
        concrete_awaitable = ConcreteAwaitable(source, curr_type, [])
        ops = (*self.el_ops.el_ops, *other.el_ops) if kind == "Subscribe" else self.el_ops.el_ops

        if isinstance(source, Subject) or isinstance(source, Observable):
            observable_chain = source
            for op in ops if kind == "Subscribe" else ops[:-1]:
                observable_chain = _op_to_functions[op[0]][0](observable_chain,  *op[1])

                curr_type = VT.Awaitable[_op_to_functions[op[0]][1](op,  *absorbed(curr_type))]
                concrete_awaitable = ConcreteAwaitable(observable_chain, curr_type, concrete_awaitable.chain)

            if kind == "Subscribe":
                from .logger import log
                observable_chain.pipe(
                    rxops.do_action(on_next=other.func),
                    rxops.do_action(on_error=lambda e: log.error(f"Error caught in rx stream", exc_info=e)),
                    rxops.retry(),
                ).subscribe()
            else:
                observable_chain = observable_chain.pipe(rxops.to_list())
                def type_assertion_wrapper(el):
                    if isinstance(el, list) and absorbed(curr_type)[0] != "List":
                        return ops[-1][1][0](only(el))
                    return ops[-1][1][0](el)
                observable_chain.subscribe(type_assertion_wrapper)

            concrete_awaitable = ConcreteAwaitable(observable_chain, *absorbed(curr_type), concrete_awaitable.chain)
            return observable_chain 
        
        raise NotImplementedError(f"Awaitable evalation not implemented with {type(source)}")

    def on_next(self, x):
        raise NotImplementedError()
        
    def on_complete(self):
        raise NotImplementedError()

class ForEachingOp:
    def __init__(self, other: OpLike):
        self.el_ops = other.el_ops
        self.func = None

    def __repr__(self):
        return f"ForEachingOp(func={self.func}, {op_chain_pretty_print(self.el_ops)})"
        
    def __ror__(self, other):
        if isinstance(other, ZefOp):
            new_zo = ZefOp( (*other.el_ops, *self.el_ops) )
            res = ForEachingOp(new_zo)
            res.func = self.func
            return res
        if isinstance(other, TraversableABC):
            res = ForEachingOp(ZefOp(((RT.TmpZefOp, (other,)), )))
            res.func = self.func
            return res
        if is_supported_value(other) or is_supported_zef_value(other):
            return LazyValue(other) | self
        raise TypeError(f'We should not have landed here. Value passed {other} of type {type(other)}.')
    
    def __or__(self, other):
        if isinstance(other, ZefOp):
            res = ForEachingOp(other)
            res.func = self.func
            return res
        if isinstance(other, LazyValue):            
            raise NotImplementedError()
        return NotImplemented

    def __gt__(self, other):
        if is_supported_zef_value(other): return LazyValue(other) < self
        
    def __lt__(self, other):
        if is_supported_zef_value(other): return LazyValue(other) > self

    def __getitem__(self, func):
        assert func is not None, "for_each requires a function"
        new_for_each = ForEachingOp(self)
        new_for_each.func = func
        return new_for_each
OpLike.register(ForEachingOp)


class SubscribingOp:
    def __init__(self, op: ZefOp):
        self.el_ops = op.el_ops     # a tuple: each element representing an elementary operator
        self.func = None

    def __repr__(self):
        return f"SubscribingOp(func={self.func}, {op_chain_pretty_print(self.el_ops)})"

    def __ror__(self, other):
        if isinstance(other, ZefOp):
            res = SubscribingOp(other)
            res.func = self.func
            return res

        if isinstance(other, TraversableABC):
            res = SubscribingOp(ZefOp(((RT.TmpZefOp, (other,)), )))
            res.func = self.func
            return res
            
        if isinstance(other, ZefRef):
            # check that this is an AET and allow subscribing natively
            raise NotImplemented
            
        raise TypeError(f'We should not have landed here. Value passed {other} of type {type(other)}.')
            
    def __or__(self, other):
        if isinstance(other, ZefOp):
            res =  SubscribingOp(other)
            res.func = self.func
            return res
        if isinstance(other, LazyValue):            
            raise NotImplementedError()
        return NotImplemented

    def __getitem__(self, func):
        new_op = SubscribingOp(ZefOp(self.el_ops))
        new_op.func =  func
        return new_op
OpLike.register(SubscribingOp)


class LazyValue:
    def __init__(self, arg):
        # if type(arg) == LazyValue:
        #     # Create a copy
        #     self.initial_val = arg.initial_val
        #     self.el_ops = arg.el_ops
        # else:
            # Allow LazyValue to encapsulate a LazyValue, if something breaks uncomment above part
            self.initial_val = arg
            self.el_ops = ()
    
    def __repr__(self):
        if type(self.el_ops) in {CollectingOp, ForEachingOp} or len(self.el_ops) > 0:
            return f"LazyValue({self.initial_val} | {self.el_ops})"
        return f"LazyValue({self.initial_val})"

    def __or__(self, other):
        if not isinstance(other, OpLike):
            return NotImplemented

        if len(self.el_ops) > 0:
            res = self.el_ops | other
        else:
            res = other
        
        res_lazyval = LazyValue(self.initial_val)
        res_lazyval.el_ops = res
        if should_trigger_eval(res_lazyval): return evaluate_lazy_value(res_lazyval)
        return res_lazyval

    def lt_gt_lshift_rshift_behavior(self, other, rt):
        import inspect
        frame = inspect.currentframe().f_back.f_back   
        other_id = id(other)

        if len(self.el_ops) > 0:
            if rt == RT.OutOld:
                res = self.el_ops > other
            elif rt == RT.InOld:
                res = self.el_ops < other
            elif rt == RT.InInOld:
                res = self.el_ops << other
            elif rt == RT.OutOutOld:
                res = self.el_ops >> other
        else:
            if isinstance(other,ZefOp) and is_tmpzefop(other):
                other_id = id(other.el_ops[0][1][0])
                res = ZefOp((*unpack_tmpzefop(rt, other.el_ops),))
            elif isinstance(other, ForEachingOp) or isinstance(other, CollectingOp):
                if is_tmpzefop(other):
                    res = type(other)(ZefOp((*unpack_tmpzefop(rt, other.el_ops), )))
                else:
                    res = type(other)(ZefOp((*unpack_ops(rt, other.el_ops),))) 
                if type(other) in { ForEachingOp}: 
                    res.func = other.func
                    other.func = None
            elif isinstance(other, TraversableABC):
                res = ZefOp(((rt, (other,)), ))
            else:
                res = ZefOp((*unpack_ops(rt, other.el_ops),))
        
        res_lazyval = LazyValue(self.initial_val)
        res_lazyval.el_ops = res
        if rt == RT.OutOld or rt == RT.InOld:
            # We are caching with as many line_nos as possible due to the fact that we don't how deep the call stack is to arrive to this point.
            while frame:_terrible_global_state[(frame.f_lineno, other_id)] = res_lazyval;frame = frame.f_back
        if should_trigger_eval(res_lazyval): return evaluate_lazy_value(res_lazyval)
        return res_lazyval

    def __gt__(self, other):
        return self.lt_gt_lshift_rshift_behavior(other, RT.OutOld)

    def __lt__(self, other):
        return self.lt_gt_lshift_rshift_behavior(other, RT.InOld)

    def __lshift__(self, other):
        return self.lt_gt_lshift_rshift_behavior(other, RT.InInOld) 

    def __rshift__(self, other):
        return self.lt_gt_lshift_rshift_behavior(other, RT.OutOutOld) 

    def __le__(self, value):
        from ._ops import assign_value
        return self | assign_value[value]

    def __iter__(self):
        return iter(self.evaluate(unpack_generator = False))

    def __call__(self, *args):
        # x = [1,2,3] | map[...] | last ; x()
        if len(args) == 0:
            return self | CollectingOp(ZefOp(((RT.Collect, ()),)))
        
        # x = [1,2,3] | map[...] | last ; x([1,2,3])
        if len(args) > 0:
            raise Exception(f"Cannot call a LazyValue with args passed!")

    # For avoiding common mistakes
    def __eq__(self, other):
        if isinstance(other, LazyValue):
            return self.initial_val == other.initial_val and self.el_ops == other.el_ops

        import logging
        logging.warning("A LazyValue has been compared with == or !=. This is likely a mistake, and requires an additional '| collect'")
        import traceback
        traceback.print_stack()
        return NotImplemented

    def __hash__(self):
        return hash((self.initial_val, self.el_ops))


    def __bool__(self):
        if getattr(self, "_allow_bool", False):
            return True
        raise Exception("Shouldn't cast LazyValue to bool (this may change in the future to automatic evaluation)")

    def evaluate(self, unpack_generator = True):
        from .op_implementations.dispatch_dictionary import _op_to_functions
        curr_value = self.initial_val
        
        # TODO: Type info has been disabled due to typespecing of long lists/sets/dicts which consumes
        # time to get the specific contained type (#ref:type_spec_iterable). For now the evaluation engine
        # doesn't depend on the type system. create_type_info could still be externally called.
        
        # primary_type_info = False
        # try:
        #     type_info = create_type_info(self)
        #     primary_type_info = True
        # except:
        #     warnings.warn("Failed to create type info using primary method. Falling back to backup type info!")
        # if not primary_type_info: back_up_type_info = [type_spec(curr_value)]

        for op in self.el_ops.el_ops: 
            if op[0] == RT.Collect: continue
            if op[0] == RT.Run:  
                from .fx.fx_types import _Effect_Class
                if isinstance(curr_value, _Effect_Class): 
                    curr_value = _op_to_functions[op[0]][0](curr_value)
                elif len(op[1]) > 1: # i.e run[impure_func]
                    curr_value = op[1][1](curr_value)
                else:
                    raise NotImplementedError(f"only effects or nullary functions can be passed to 'run' to be executed in the imperative shell. Received {curr_value}")
                break
            

            # TODO Throw a panic here
            try:
                curr_value = _op_to_functions[op[0]][0](curr_value,  *op[1])
            except Exception as e:
                raise RuntimeError(f"An Exception occured while evaluating '{self.initial_val} | {ZefOp(self.el_ops.el_ops)}'.\nThe exception occured while calling {op[0]} with ({curr_value,  *op[1]}). The exception was {e}") from e
            
            from .error import _ErrorType
            if isinstance(curr_value, _ErrorType): raise RuntimeError(f"The evaluation engine returned an Error from '{ZefOp((op,))}'.\nThe error was {curr_value}")
            
            # if not primary_type_info: 
            #     curr_type, curr_value = find_type_of_current_value(curr_value)
            #     back_up_type_info.append(curr_type)
        # self.type_info = type_info if primary_type_info else back_up_type_info
        
        if unpack_generator and (isinstance(curr_value, Iterator) or isinstance(curr_value, Generator)):
            return [i for i in curr_value]
        return curr_value      

def find_type_of_current_value(curr_value):
    if isinstance(curr_value, Iterator) or isinstance(curr_value, Generator): 
        try:
            first_val = next(curr_value)
            def output_generator():
                yield first_val
                yield from curr_value
            return type_spec(first_val), output_generator()
        except Exception:
            warnings.warn("Cannot infer data type from an empty generator. Using Any instead!")
            return VT.Any, curr_value 
    else: 
        return type_spec(curr_value), curr_value                                                                                     

#   _____                _                _    _                   _____                _              
#  | ____|__   __  __ _ | | _   _   __ _ | |_ (_)  ___   _ __     | ____| _ __    __ _ (_) _ __    ___ 
#  |  _|  \ \ / / / _` || || | | | / _` || __|| | / _ \ | '_ \    |  _|  | '_ \  / _` || || '_ \  / _ \
#  | |___  \ V / | (_| || || |_| || (_| || |_ | || (_) || | | |   | |___ | | | || (_| || || | | ||  __/
#  |_____|  \_/   \__,_||_| \__,_| \__,_| \__||_| \___/ |_| |_|   |_____||_| |_| \__, ||_||_| |_| \___|
#                                                                                |___/                 

def is_evaluating_run(op):
    if isinstance(op, ZefOp) and op.el_ops[-1][0] == RT.Run and op.el_ops[-1][1][0] == evaluating:
        return True
    return False

def should_trigger_eval(value_maybe):
    if isinstance(value_maybe, LazyValue):
        if is_evaluating_run(value_maybe.el_ops): return True
        if isinstance(value_maybe.el_ops, CollectingOp) or isinstance(value_maybe.el_ops, ForEachingOp):
            return True
        else:
            for op in value_maybe.el_ops.el_ops: # These are the Ops of the nested ZefOp
                if op[0] in {RT.Collect}: return True 
    
    elif isinstance(value_maybe, Awaitable):
        raise NotImplementedError("Evaluating isn't implemented yet for Awaitables!")
    
    return False

# ----LazyValue evaluation-----
def evaluate_lazy_value_with_curried_op(lazyval: LazyValue) -> LazyValue:
    # Testcase: LazyValue([1,2,3]) | map[mapper] | collect | map[mapper] 
    if isinstance(lazyval.el_ops, ForEachingOp):
        if lazyval.el_ops.func is None: raise Exception("The curried ForeachingOp doesn't have a defined function.")
        res = [lazyval.el_ops.func(x) for x in lazyval]
    elif isinstance(lazyval.el_ops, CollectingOp):
        res = lazyval.evaluate()
    elif is_evaluating_run(lazyval.el_ops):
        res = lazyval.evaluate()
    else:
        raise NotImplementedError('No supported type for trigger_eval!')
    return res

def evaluate_lazy_value_with_zefop(lazyval: LazyValue) -> LazyValue:
    # Testcase: LazyValue([1,2,3]) | map[mapper] | collect >> filter[mapper]
    ops = lazyval.el_ops.el_ops
    for idx, op in enumerate(ops):
        if op[0] in {RT.Collect}: break
    tmp_ops = ops[idx + 1:] if idx + 1 < len(ops) else ()
    if ops[idx][0] == RT.Collect:
        tmp_op = CollectingOp(ZefOp(ops[:idx]))
    else:
        raise NotImplementedError

    new_inner = LazyValue(lazyval)
    new_inner.el_ops = tmp_op
    new_lazyval = LazyValue(evaluate_lazy_value_with_curried_op(new_inner))
    new_lazyval.el_ops = ZefOp(tmp_ops)
    return new_lazyval

def evaluate_lazy_value(lazyval: LazyValue) -> LazyValue:
    if type(lazyval.el_ops) in {CollectingOp, ForEachingOp} or is_evaluating_run(lazyval.el_ops):
        return evaluate_lazy_value_with_curried_op(lazyval)
    elif isinstance(lazyval.el_ops, ZefOp):
        return evaluate_lazy_value_with_zefop(lazyval)
    
    raise NotImplementedError(f"evaluate_lazy_value not implemented with {type(lazyval.el_ops)}")
        
    

# ---- Awaitable evaluation -----
def evaluate_Awaitable(Awaitable: Awaitable) -> Awaitable:
    raise NotImplementedError("evaluate_Awaitable")


#   _____  _             _     _____                         ___          __        
#  |  ___|(_) _ __    __| |   |_   _| _   _  _ __    ___    |_ _| _ __   / _|  ___  
#  | |_   | || '_ \  / _` |     | |  | | | || '_ \  / _ \    | | | '_ \ | |_  / _ \ 
#  |  _|  | || | | || (_| |     | |  | |_| || |_) ||  __/    | | | | | ||  _|| (_) |
#  |_|    |_||_| |_| \__,_|     |_|   \__, || .__/  \___|   |___||_| |_||_|   \___/ 
#                                     |___/ |_|                                     

def get_compiled_zeffunction(fct):
    from .. import zef_functions
    try:   
        fct = zef_functions._local_compiled_zef_functions[zef_functions.time_resolved_hashable(fct)]
    except KeyError:
        fct = zef_functions.compile_zef_function(fct)
    return fct

def type_spec_iterable(obj, vt_type):
    # TODO: this causes slowness with large lists i.e list(range(10**1000))
    try:
        tps = set(type_spec(e) for e in obj)
        if len(tps) == 1:
            return vt_type[next(iter(tps))]
        else:
            return vt_type[VT.Any]
    except:
        return vt_type[VT.Any]
        
def type_spec_dict(obj):
    tps = set((type_spec(k), type_spec(v)) for (k,v) in obj.items())
    keytypes = set(k for (k, v) in tps)
    valtypes =  set(v for (k, v) in tps)
    kt = next(iter(keytypes)) if len(keytypes) == 1 else VT.Any
    vt = next(iter(valtypes)) if len(valtypes) == 1 else VT.Any
    res = VT.Dict[kt, vt]
    return res

def type_spec_tuple(obj):
    new_tup = VT.Record
    if len(obj) > 0:
        return new_tup[type_spec(obj[0])]
    else:
        return new_tup[VT.Any]

def type_spec(obj, no_type_casting = False):
    from .VT import ValueType_
    from .fx.fx_types import _Effect_Class
    from . import GraphSlice
    from rx.subject import Subject
    from rx.core import Observable
    if isinstance(obj, ValueType_):               return obj
    if isinstance(obj, type) or no_type_casting: t = obj
    else:                                        t = type(obj)
    res = {
        int:                        VT.Int,
        str:                        VT.String,
        bool:                       VT.Bool,
        float:                      VT.Float,
        type(None):                 VT.Nil,
        list:                       lambda o: type_spec_iterable(o, VT.List),
        set:                        lambda o: type_spec_iterable(o,  VT.Set),
        dict:                       type_spec_dict,
        tuple:                      type_spec_tuple,
        ZefRef:                     VT.ZefRef,
        ZefRefs:                    VT.ZefRefs,
        ZefRefss:                   VT.ZefRefss,
        EZefRef:                    VT.EZefRef,
        EZefRefs:                   VT.EZefRefs,
        EZefRefss:                  VT.EZefRefss,
        Graph:                      VT.Graph,
        Time:                       VT.Time,  
        Subject:                    VT.Awaitable,
        Observable:                 VT.Awaitable,
        _Effect_Class:              VT.Effect,
        GraphSlice:                 VT.GraphSlice,
    }.get(t, lambda o: ValueType_(type(o).__name__, 0))
    try:
        return res if str(res) in dir(VT) else res(obj)
    except Exception as e:
        raise Exception(f"An error happened in type_spec for arg {obj}") from e


def create_type_info(lazyval) -> list:
    from .op_implementations.dispatch_dictionary import _op_to_functions
    curr_type = type_spec(lazyval.initial_val)
    type_transformation = [curr_type]
    for op in lazyval.el_ops.el_ops:
        if op[0] in {RT.Collect , RT. RT.Run}:
            pass
        elif op[0] in _op_to_functions:
            curr_type = _op_to_functions[op[0]][1](op, curr_type)
        else:
            raise NotImplementedError(f"{op[0]} isn't handled in create_type_info")
        
        type_transformation.append(curr_type)
    return type_transformation


#   ____   _____     ____          _          _      _               
#  |  _ \ |_   _|   |  _ \   __ _ | |_   ___ | |__  (_) _ __    __ _ 
#  | |_) |  | |     | |_) | / _` || __| / __|| '_ \ | || '_ \  / _` |
#  |  _ <   | |     |  __/ | (_| || |_ | (__ | | | || || | | || (_| |
#  |_| \_\  |_|     |_|     \__,_| \__| \___||_| |_||_||_| |_| \__, |
#                                                              |___/ 

from ..core import ZefRefs, EZefRef, EZefRefs

def new__rshift__(self, arg):
    # promote RT or BT then rshift
    return ZefOp(((RT.TmpZefOp, (self,)), )) >> arg

def new__rrshift__(self, arg):
  # promote RT or BT then rshift with arg first
  return arg >> ZefOp(((RT.TmpZefOp, (self,)), )) 

RelationType.__rshift__ = new__rshift__
RelationType.__rrshift__ = new__rrshift__
internals.BlobType.__rshift__ = new__rshift__
internals.BlobType.__rrshift__ = new__rrshift__

# monkey patch << for RT
def new__lshift__(self, arg):
    # promote RT or BT then rshift
    return ZefOp(((RT.TmpZefOp, (self,)), )) << arg

def new__rlshift__(self, arg):
  # promote RT or BT then rshift with arg first
  return arg << ZefOp(((RT.TmpZefOp, (self,)), )) 

RelationType.__lshift__ = new__lshift__
RelationType.__rlshift__ = new__rlshift__
internals.BlobType.__lshift__ = new__lshift__
internals.BlobType.__rlshift__ = new__rlshift__


def rt_lt_gt_behavior(self, arg, rt):
    # Because > , < aren't defined for ZefRef the RT binds but with the reverse of the operator i.e > becomes <
    if isinstance(arg, ZefRef) or isinstance(arg, ZefRefs) or isinstance(arg, EZefRef) or isinstance(arg, EZefRefs):
        if rt == RT.InOld: return arg > ZefOp(((RT.TmpZefOp, (self,)), )) 
        else: return arg < ZefOp(((RT.TmpZefOp, (self,)), )) 
    elif isinstance(self, TraversableABC) and (isinstance(arg, TraversableABC) or isinstance(arg, ZefOp)):
        import inspect
        frame = inspect.currentframe().f_back.f_back
        line_no = frame.f_lineno     
        used_id = id(self)
        is_non_op = False
        if isinstance(arg, ZefOp):      
            # This is the case if arg is of instance TmpZefOp so we need to unwrap it
            if is_tmpzefop(arg): 
                arg_ops = unpack_tmpzefop(rt, arg.el_ops)
            # This handles all other ZefOp cases including L zefop
            else: 
                 arg_ops = unpack_ops(rt, arg.el_ops)

            if (line_no, used_id) in _terrible_global_state:
                cached = _terrible_global_state[(line_no, used_id)]
                if isinstance(cached, LazyValue) or isinstance(cached, Awaitable):
                    res = ZefOp((*(cached.el_ops.el_ops), *arg_ops))
                    if isinstance(cached, LazyValue): cached_clone = type(cached)(cached.initial_val)
                    else: cached_clone = type(cached)(cached.stream_ezefref, cached.pushable, cached.unwrapping)
                    cached_clone.el_ops = res
                    res = cached_clone
                    is_non_op = True
                else:
                    res = ZefOp((*(_terrible_global_state[(line_no, used_id)]), *arg_ops))
            else:
                res = ZefOp(((RT.TmpZefOp, (self,)), *arg_ops))


        # Other is a RelationType and self is cached previously so use previous result
        elif(line_no, used_id) in _terrible_global_state:
            cached = _terrible_global_state[(line_no, used_id)]
            if isinstance(cached, LazyValue) or isinstance(cached, Awaitable):
                res = ZefOp(( *(cached.el_ops.el_ops), (rt, (arg,))) )
                if isinstance(cached, LazyValue): cached_clone = type(cached)(cached.initial_val)
                else: cached_clone = type(cached)(cached.stream_ezefref, cached.pushable, cached.unwrapping)
                cached_clone.el_ops = res
                res = cached_clone
                is_non_op = True
            else:
                res = ZefOp(( *(_terrible_global_state[(line_no, used_id)]), (rt, (arg,))) )
                
        # Other is a RelationType and there is no previous result
        else:
            res = ZefOp(((RT.TmpZefOp, (self,)), (rt, (arg,)))) 
        _terrible_global_state[(line_no, id(arg))] = res if is_non_op else res.el_ops
        if isinstance(res, LazyValue)  and should_trigger_eval(res):
            return evaluate_lazy_value(res)
        return res

    elif isinstance(self, TraversableABC) and (isinstance(arg, CollectingOp) or isinstance(arg, ForEachingOp) or isinstance(arg, SubscribingOp)):
        import inspect
        frame = inspect.currentframe().f_back.f_back
        line_no = frame.f_lineno     
        used_id = id(self)

        if is_tmpzefop(arg): 
            args = (*unpack_tmpzefop(rt, arg.el_ops),)
        else:
            args = (*unpack_ops(rt, arg.el_ops),)
            
        if(line_no, used_id) in _terrible_global_state:
            cached = _terrible_global_state[(line_no, used_id)]
            # Situation where RT was cached as an LazyValue or Awaitable and the Arg is Collecting/ForEachingOp
            if isinstance(cached, LazyValue) or isinstance(cached, Awaitable):
                res = type(arg)(ZefOp(( *(cached.el_ops.el_ops), *args)))
                if isinstance(cached, LazyValue): cached_clone = type(cached)(cached.initial_val)
                else: cached_clone = type(cached)(cached.stream_ezefref, cached.pushable, cached.unwrapping)
                cached_clone.el_ops = res
                res = cached_clone
            else:
                res = type(arg)(ZefOp(( *(_terrible_global_state[(line_no, used_id)]), *args)))
        else:
            res = type(arg)(ZefOp(((RT.TmpZefOp, (self,)), *args)))
        
        # Pass the curried function to the new res and reset the previously set op
        if type(arg) in {SubscribingOp, ForEachingOp}: 
            if type(res) in {LazyValue, Awaitable}:
                res.el_ops.func = arg.func
            else:
                res.func = arg.func
            arg.func = None

        _terrible_global_state[(line_no, id(arg))] = res 
        if isinstance(res, LazyValue)  and should_trigger_eval(res):
            return evaluate_lazy_value(res)

        return res

    print(f'unsupported operand type(s) for {">" if rt == RT.OutOld else "<"} {self} , {arg}')
    raise NotImplemented

# monkey patch > for RT and BT (or z < RT when z has no __lt__)
old__RT_gt__ = RelationType.__gt__
def new__gt__(self, arg):
    return rt_lt_gt_behavior(self, arg, RT.OutOld)
RelationType.__gt__ = new__gt__
internals.BlobType.__gt__ = new__gt__

# monkey patch < for RT and BT (or z > RT when z has no __gt__)
old__RT_lt__ = RelationType.__lt__
def new__lt__(self, arg):
    return rt_lt_gt_behavior(self, arg, RT.InOld)
RelationType.__lt__ = new__lt__
internals.BlobType.__lt__ = new__lt__


def make_ror_promote_zefop(rt):
    return lambda rhs,lhs: lhs | ZefOp(((rt, (rhs,)), ))

from functools import partial
internals.EntityTypeStruct.__ror__ = make_ror_promote_zefop(RT.ET)
internals.RelationTypeStruct.__ror__ = make_ror_promote_zefop(RT.RT)
internals.AtomicEntityTypeStruct.__ror__ = make_ror_promote_zefop(RT.AET)
internals.BlobTypeStruct.__ror__ = make_ror_promote_zefop(RT.BT)



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .._core import RT
from .implementation_typing_functions import * 
from .yo import yo_implementation, yo_type_info
from .graphviz import graphviz_imp, graphviz_tp
from .func import call_implementation, call_type_info
from .func import unpack_implementation, unpack_type_info

_op_to_functions = {
        RT.ApplyFunctions: (apply_functions_imp, None),
        RT.Map:            (map_implementation, map_type_info),
        RT.Reduce:         (reduce_implementation, reduce_type_info),
        RT.Scan:           (scan_implementation, scan_type_info),
        RT.GroupBy:        (group_by_implementation, group_by_type_info),
        RT.Transpose:      (transpose_imp, transpose_tp),
        RT.Frequencies:    (frequencies_imp, frequencies_tp),
        RT.Iterate:        (iterate_implementation, iterate_type_info),
        RT.Identity:       (identity_implementation, identity_type_info),
        RT.Length:         (length_implementation, length_type_info),
        RT.Take :          (take_implementation, take_type_info),
        RT.TakeWhile:      (take_while_imp, take_while_tp),
        RT.TakeWhilePair:  (take_while_pair_imp, take_while_pair_tp),
        RT.TakeUntil:      (take_until_imp, take_until_tp),
        RT.SkipWhile:      (skip_while_imp, skip_while_tp),
        RT.Drop :          (drop_imp, drop_tp),
        RT.Skip :          (skip_imp, skip_tp),
        RT.Nth:            (nth_implementation, nth_type_info),
        RT.Filter:         (filter_implementation, filter_type_info),
        RT.SelectKeys:     (select_keys_imp, None),
        RT.Modulo:         (modulo_imp, None),
        RT.SelectByField:  (select_by_field_imp, select_by_field_tp),
        RT.First:          (first_imp, first_tp),
        RT.Second:         (second_imp, second_tp),
        RT.Last:           (last_imp, last_tp),
        RT.Single:         (single_imp, single_tp),
        RT.SingleOr:       (single_or_imp, single_or_tp),
        RT.Zip:            (zip_imp, zip_tp),
        RT.Concat:         (concat_implementation, concat_type_info),
        RT.Prepend:        (prepend_imp, prepend_tp),
        RT.Append:         (append_imp, append_tp),
        RT.Interleave:     (interleave_imp, interleave_tp),
        RT.InterleaveLongest: (interleave_longest_imp, interleave_longest_tp),
        RT.Sort:            (sort_implementation, sort_type_info),
        RT.Now:             (now_implementation, now_type_info),
        RT.Merged:          (merged_imp, merged_tp),
        RT.Affected:        (affected_imp, affected_tp),
        RT.ValueAssigned:   (value_assigned_imp, value_assigned_tp),
        RT.Instantiated:    (instantiated_imp, instantiated_tp),
        RT.Terminated:      (terminated_imp, terminated_tp),
        RT.Events:          (events_imp, events_tp),
        RT.ToDelegate:      (to_delegate_implementation, to_delegate_type_info),
        RT.DelegateOf:      (delegate_of_implementation, delegate_of_type_info),
        
        RT.In:              (In_imp, None),
        RT.Ins:             (Ins_imp, None),
        RT.Out:             (Out_imp, None),
        RT.Outs:            (Outs_imp, None),
        RT.InRel:           (in_rel_imp, None),
        RT.InRels:          (in_rels_imp, None),
        RT.OutRel:          (out_rel_imp, None),
        RT.OutRels:         (out_rels_imp, None),

        RT.InOld:              (InOld_implementation, None),
        RT.InInOld:            (InInOld_implementation, None),
        RT.OutOld:             (OutOld_implementation, None),
        RT.OutOutOld:          (OutOutOld_implementation, None),
        
        RT.InsOld:          (ins_implementation_old, ins_type_info),
        RT.OutsOld:         (outs_implementation_old, outs_type_info),
        RT.InsAndOutsOld:   (ins_and_outs_implementation_old, ins_and_outs_type_info),
        RT.Source:          (source_implementation, source_type_info),
        RT.Target:          (target_implementation, target_type_info),
        RT.Value:           (value_implementation, value_type_info),
        RT.Time:            (time_implementation, time_type_info),
        RT.TimeSlice:       (time_slice_implementation, time_slice_type_info),
        RT.Tx:              (tx_imp, tx_tp),
        RT.NextTX:          (next_tx_imp, next_tx_tp),
        RT.PreviousTX:      (previous_tx_imp, previous_tx_tp),
        RT.Tx:              (tx_imp, tx_tp),
        RT.InstantiationTx: (instantiation_tx_implementation, instantiation_tx_type_info),                              # TODO: retire
        RT.TerminationTx:   (termination_tx_implementation, termination_tx_type_info),                            # TODO: retire
        RT.Instances:       (instances_implementation, instances_type_info),
        RT.Uid:             (uid_implementation, uid_type_info),
        RT.Frame:           (frame_imp, frame_tp),
        RT.BaseUid:         (base_uid_implementation, base_uid_type_info),
        RT.OriginUid:       (origin_uid_imp, origin_uid_tp),
        RT.OriginRAE:       (origin_rae_imp, origin_rae_tp),
        RT.ExistsAt:        (exists_at_implementation, exists_at_type_info),
        RT.IsZefRefPromotable:(is_zefref_promotable_implementation, is_zefref_promotable_type_info),
        RT.InFrame:         (in_frame_imp, in_frame_tp),
        RT.ToGraphSlice:    (to_graph_slice_imp, to_graph_slice_tp),
        RT.ToTx:            (to_tx_imp, to_tx_tp),
        RT.TimeTravel:      (time_travel_imp, time_travel_tp),
        RT.ToEZefRef:       (to_ezefref_implementation, to_ezefref_type_info),
        RT.O:               (o_implementation, o_type_info),
        RT.L:               (l_implementation, l_type_info),
        RT.Terminate:       (terminate_implementation, terminate_type_info),
        RT.AssignValue:     (assign_value_imp, assign_value_tp),
        RT.ET:              (ET_implementation, ET_type_info),
        RT.RT:              (RT_implementation, RT_type_info),
        RT.AET:             (AET_implementation, AET_type_info),
        RT.BT:              (BT_implementation, BT_type_info),
        RT.FillOrAttach:    (fill_or_attach_implementation, fill_or_attach_type_info),
        RT.Assert:          (assert_implementation, assert_type_info),
        RT.HasOut:          (hasout_implementation, hasout_type_info),
        RT.HasIn:           (hasin_implementation, hasin_type_info),
        RT.Run:             (run_effect_implementation, None),
        RT.Tap:             (tap_imp, tap_tp),
        RT.Push:            (push_imp, push_tp),
        RT.IsA:             (is_a_implementation, is_a_type_info),
        RT.IsRepresentedAs: (is_represented_as_implementation, is_represented_as_type_info),
        RT.ZefType:         (zef_type_imp, None),
        RT.HasRelation:     (has_relation_implementation, has_relation_type_info),
        RT.Relation:        (relation_implementation, relation_type_info),
        RT.Relations:       (relations_implementation, relations_type_info),
        RT.Chunk:           (chunk_imp, chunk_tp),
        RT.Sliding:         (sliding_imp, sliding_tp),
        RT.Stride:          (stride_imp, stride_tp),
        RT.Insert:          (insert_imp, insert_tp),
        RT.InsertInto:      (insert_into_imp, None),
        RT.ReverseArgs:     (reverse_args_imp, None),
        RT.Remove:          (remove_imp, remove_tp),
        RT.Get:             (get_imp, get_tp),
        RT.GetField:        (get_field_imp, get_field_tp),
        RT.Enumerate:       (enumerate_imp, enumerate_tp),
        RT.Items:           (items_imp, items_tp),
        RT.Values:          (values_imp, values_tp),
        RT.Keys:            (keys_imp, keys_tp),
        RT.Reverse:         (reverse_imp, reverse_tp),
        RT.RaeType:         (rae_type_implementation, rae_type_type_info),
        RT.AbstractType:    (abstract_type_implementation, abstract_type_type_info),
        RT.Root:            (root_imp, root_tp),
        RT.Blobs:           (blobs_imp, blobs_tp),
        RT.Z:               (Z_imp, Z_tp),
        RT.Docstring:       (docstring_imp, None),
        
        
        RT.GetIn:           (get_in_imp, get_in_tp),
        RT.InsertIn:        (insert_in_imp, insert_in_tp),
        RT.Update:          (update_imp, None),
        RT.UpdateIn:        (update_in_imp, update_in_tp),
        RT.UpdateAt:        (update_at_imp, update_at_tp),
        RT.RemoveIn:        (remove_in_imp, remove_in_tp),
        RT.RemoveAt:        (remove_at_imp, None),
        RT.Merge:           (merge_imp, merge_tp),
        RT.MergeWith:       (merge_with_imp, merge_with_tp),
        RT.IntToAlpha:      (int_to_alpha_imp, None),
        RT.PermuteTo:       (permute_to_imp, None),
                
        RT.Cycle:           (cycle_imp, cycle_tp),
        RT.Repeat:          (repeat_imp, repeat_tp),
        RT.Contains:        (contains_imp, contains_tp),
        RT.ContainedIn:     (contained_in_imp, contained_in_tp),
        # RT.Skip:            (skip_imp, skip_tp),
        RT.All:             (all_imp, all_tp),
        RT.Any:             (any_imp, any_tp),
        RT.Join:            (join_imp, join_tp),
        RT.Trim:            (trim_imp, trim_tp),
        RT.TrimLeft:        (trim_left_imp, trim_left_tp),
        RT.TrimRight:       (trim_right_imp, trim_right_tp),
        RT.Yo:              (yo_implementation, yo_type_info),
        
        RT.Sign:            (sign_imp, sign_tp),
        RT.IfThenElse:      (if_then_else_imp, if_then_else_tp),
        RT.IfThenElseApply:      (if_then_else_apply_imp, if_then_else_apply_tp),
        RT.Attempt:         (attempt_imp, attempt_tp),
        RT.Bypass:          (bypass_imp, bypass_tp),
        RT.Pattern:         (pattern_imp, pattern_tp),
        RT.Replace:         (replace_imp, replace_tp),
        RT.Distinct:        (distinct_imp, distinct_tp),
        RT.DistinctBy:      (distinct_by_imp, distinct_by_tp),
        RT.IsDistinct:      (is_distinct_imp, is_distinct_tp),
        RT.IsDistinctBy:    (is_distinct_by_imp, is_distinct_by_tp),
        RT.Shuffle:         (shuffle_imp, shuffle_tp),
        RT.Slice:           (slice_imp, slice_tp),
        RT.Split:           (split_imp, split_tp),
        RT.Graphviz:        (graphviz_imp, graphviz_tp),
        
        RT.Always:              (always_imp, always_tp),
        
        RT.WithoutAbsorbed:     (without_absorbed_imp, None),
        RT.Absorbed:            (absorbed_imp, None),
        
        RT.CartesianProduct:    (cartesian_product_imp, cartesian_product_tp),
        RT.Permutations:        (permutations_imp, permutations_tp),
        RT.Combinations:        (combinations_imp, combinations_tp),
        RT.Add:                 (add_imp, add_tp),
        RT.Subtract:            (subtract_imp, subtract_tp),
        RT.Multiply:            (multiply_imp, multiply_tp),
        RT.Divide:              (divide_imp, divide_tp),
        RT.Mean:                (mean_imp, mean_tp),
        RT.Variance:            (variance_imp, variance_tp),
        RT.Power:               (power_imp, power_tp),
        RT.Exponential:         (exponential_imp, exponential_tp),
        RT.Logarithm:           (logarithm_imp, logarithm_tp),
        RT.Max:                 (max_imp, max_tp),
        RT.Min:                 (min_imp, min_tp),
        RT.MaxBy:               (max_by_imp, max_by_tp),
        RT.MinBy:               (min_by_imp, min_by_tp),
        RT.Equals:              (equals_imp, equals_tp),
        RT.LargerThan:          (greater_than_imp, greater_than_tp),
        RT.LessThan:            (less_than_imp, less_than_tp),
        RT.GreaterThanOrEqual:  (greater_than_or_equal_imp, greater_than_or_equal_to),
        RT.LessThanOrEqual:     (less_than_or_equal_imp, less_than_or_equal_to),
        RT.Not:                 (not_imp, not_tp),
        RT.And:                 (and_imp, and_tp),
        RT.Or:                  (or_imp, or_tp),
        RT.Xor:                 (xor_imp, xor_tp),
        RT.Peel:                (peel_imp, peel_tp),
        RT.Match:               (match_imp, match_tp),
        RT.MatchApply:          (match_apply_imp, match_apply_tp),

        RT.Sync:                (sync_imp, sync_tp),
        RT.Tag:                 (tag_imp, tag_tp),
        RT.Untag:               (untag_imp, untag_tp),        
        RT.ToClipboard:         (to_clipboard_imp, to_clipboard_tp),
        RT.FromClipboard:       (from_clipboard_imp, from_clipboard_tp),
        RT.TextArt:             (text_art_imp, text_art_tp),

        RT.ToJSON:             (to_json_imp, to_json_tp), 
        RT.FromJSON:           (from_json_imp, from_json_tp),

        RT.ToYaml:             (to_yaml_imp, to_yaml_tp),
        RT.FromYaml:           (from_yaml_imp, from_yaml_tp),

        RT.ToToml:              (to_toml_imp, to_toml_tp),
        RT.FromToml:            (from_toml_imp, from_toml_tp),

        RT.ToCSV:               (to_csv_imp, from_csv_tp),
        RT.FromCSV:             (from_csv_imp, from_csv_tp),

        RT.ReadFile:            (read_file_imp, read_file_tp),
        RT.LoadFile:            (load_file_imp, load_file_tp),
        RT.WriteFile:           (write_file_imp, write_file_tp),
        RT.SaveFile:            (save_file_imp, save_file_tp),

        RT.PandasToGd:          (pandas_to_gd_imp, pandas_to_gd_tp),
        
        RT.AsPipeline:          (as_pipeline_imp, as_pipeline_tp),
        RT.Inject:              (inject_imp, inject_tp),
        RT.InjectList:          (inject_list_imp, inject_list_tp),

        RT.ZasciiToAsg:         (zascii_to_asg_imp, zascii_to_asg_tp),
        RT.ZasciiToScehma:      (zascii_to_schema_imp, zascii_to_schema_tp),

        RT.ReplaceAt:           (replace_at_imp, replace_at_tp),
        RT.RandomPick:          (random_pick_imp, random_pick_tp),
        RT.PadToLength:         (pad_to_length_imp, pad_to_length_tp),
        RT.IsAlpha:             (is_alpha_imp, is_alpha_tp),
        RT.ToUpperCase:         (to_upper_case_imp, to_upper_case_tp),
        RT.ToLowerCase:         (to_lower_case_imp, to_lower_case_tp),
        RT.ToPascalCase:        (to_pascal_case_imp, to_pascal_case_tp),
        RT.ToCamelCase:         (to_camel_case_imp, to_camel_case_tp),
        RT.ToKebabCase:         (to_kebab_case_imp, to_kebab_case_tp),
        RT.ToSnakeCase:         (to_snake_case_imp, to_snake_case_tp),
        RT.ToScreamingSnakeCase:(to_screaming_snake_case_imp, to_screaming_snake_case_tp),


        RT.Call:                (call_implementation, call_type_info),
        RT.Unpack:              (unpack_implementation, unpack_type_info),

        RT.MakeRequest:        (make_request_imp, make_request_tp),
        RT.Blake3:             (blake3_imp, blake3_tp),
        RT.ToZefList:          (to_zef_list_imp, to_zef_list_tp),
        RT.Transact:           (transact_imp, transact_tp),

        RT.Function:           (function_imp, function_tp),
        RT.On:                 (on_implementation, None),
        RT.Range:              (range_imp, range_tp),
}
